from pathlib import Path
from typing import Optional

from numerous.cli.auth import get_refresh_token
from numerous.cli.job import start_local_job
from numerous.cli.repository import NumerousRepository
from numerous.cli.utils import print_load_warnings


def create_local_execution_from_repository(
    path: Optional[Path] = None,
) -> tuple[NumerousRepository, str, str]:
    path = path or Path.cwd()
    repo, load_warnings = NumerousRepository(path).load()
    print_load_warnings(load_warnings)

    if repo.remote is None or repo.scenario is None:
        raise RuntimeError(
            "Repository not configured correctly, try checking out a scenario or re-initializing."
        )

    execution_id = start_local_job(repo)
    refresh_token = get_refresh_token(repo, execution_id)

    return repo, execution_id, refresh_token



from numerous.grpc import spm_pb2, spm_pb2_grpc

from .auth import IdTokenAuth, get_id_token
from .repository import NumerousRepository
from .utils import OrganizationInterceptor, StubClient


def start_local_job(repo: NumerousRepository) -> str:
    if repo.remote is None:
        raise RuntimeError("Repository remote is not configured, use: numerous config")

    if repo.scenario is None:
        raise RuntimeError(
            "Repository scenario is not checked out, use: numerous checkout"
        )

    if repo.refresh_token is None:
        raise RuntimeError("Repository has no refresh token, use: numerous login")

    request = spm_pb2.StartLocalJobRequest(
        project_id=repo.scenario.project_id,
        scenario_id=repo.scenario.id,
        job_id=repo.remote.job_id,
    )

    id_token = get_id_token(
        repo.remote.api_url, repo.refresh_token, repo.remote.organization
    )

    with StubClient(
        repo.remote.api_url,
        spm_pb2_grpc.JobManagerStub,
        IdTokenAuth(id_token),
        [OrganizationInterceptor(repo.remote.organization)],
    ) as job_manager:
        return str(job_manager.StartLocalJob(request).execution_id)



import json
from dataclasses import dataclass
from typing import Any, Iterator, NoReturn

import grpc
from numerous.grpc import spm_pb2, spm_pb2_grpc


@dataclass
class Message:
    channel: str
    message: Any

    @staticmethod
    def from_proto(message: spm_pb2.SubscriptionMessage):
        return Message(
            channel=message.channel,
            message=json.loads(message.message),
        )


class Subscription:
    def __init__(
        self,
        project_id: str,
        scenario_id: str,
        spm_stub: spm_pb2_grpc.SPMStub,
        channel_patterns: list[str],
    ):
        self._spm_stub = spm_stub
        self._stream: Iterator[
            spm_pb2.SubscriptionMessage
        ] = self._spm_stub.SubscribeForUpdates(
            spm_pb2.Subscription(
                channel_patterns=channel_patterns,
                scenario=scenario_id,
                project_id=project_id,
            )
        )  # type: ignore[assignment]

    def _handle_cancelled_if_subscription_closed(
        self, error: grpc.RpcError
    ) -> NoReturn:
        if (
            error.code() == grpc.StatusCode.CANCELLED
            and error.details() == "Channel closed!"
        ):
            raise StopIteration

        else:
            raise

    def __iter__(self) -> "Subscription":
        return self

    def __next__(self) -> Message:
        try:
            return Message.from_proto(next(self._stream))
        except grpc.RpcError as error:
            self._handle_cancelled_if_subscription_closed(error)



import io
import sys
from logging import Logger
from pathlib import Path

import requests

from ..auth import AuthenticationError, login
from ..build_manager_client import get_build_manager_client
from ..repository import NumerousRepository, ProjectScenario, RepositoryRemote
from ..utils import (
    bold,
    cyan,
    extract_gzipped_tarball,
    green,
    print_repository_help_message,
    red,
)


def cleanup(path: Path):
    (path / NumerousRepository.FILE_NAME).unlink(True)
    (path / ".exclude").unlink(True)
    path.rmdir()


def command_clone(
    log: Logger, scenario: ProjectScenario, path: Path, remote: RepositoryRemote
) -> None:
    path = path or (Path.cwd() / remote.name)

    if path.exists():
        print(red(f"Cannot clone: {bold(path)} already exists"))
        sys.exit(1)

    repo = NumerousRepository(path)
    repo.remote = remote
    repo.scenario = scenario
    path.mkdir()

    try:
        try:
            id_token = login(repo)
        except AuthenticationError:
            print(red("Cannot clone: Login failed."))
            cleanup(path)
            sys.exit(1)

        with get_build_manager_client(
            repo.remote.api_url, id_token, repo.remote.organization
        ) as build_manager:
            updates = build_manager.get_scenario_history(repo.remote.name, scenario.id)

            if not updates:
                print(red("Cannot clone: No commit found"))
                sys.exit(1)

            update = updates[0]
            repo.snapshot = update.snapshot_id
            repo.scenario = scenario

            print(
                cyan(
                    f"Downloading snapshot {bold(repo.snapshot)} into {bold(path.absolute())}"
                )
            )
            url = build_manager.get_download_url(
                repo.remote.name,
                repo.scenario.project_id,
                repo.scenario.id,
                repo.snapshot,
            )
            response = requests.get(url, allow_redirects=True)
            extract_gzipped_tarball(io.BytesIO(response.content), path)

            repo.commit = update.commit_id
            repo.save()
            abs_path = path.absolute()
            print(
                green(
                    f"Repository {bold(repo.remote.name)} was cloned into {bold(abs_path)}"
                )
            )
            print_repository_help_message(abs_path)
    except BaseException:
        path.rmdir()
        print(red("Cannot clone: An error occured."))
        sys.exit(1)
    repo.save()



import logging
import threading
from abc import ABC, abstractmethod
from time import sleep
from typing import Any, Dict, List, Tuple

from ..numerous_client import NumerousClient, ScenarioStatus
from .configuration import ScenarioSetting

log = logging.getLogger(__name__)


class OptimizationIteration:
    def __init__(
        self,
        scenario_id: str,
        execution_id: str,
        scenario_settings: List[ScenarioSetting],
    ):
        self.scenario_id = scenario_id
        self.execution_id = execution_id
        self.scenario_settings = scenario_settings

    def __str__(self):
        settings = [
            f'{s["type"]}: {s["id"]}={s["value"]}' for s in self.scenario_settings
        ]
        return f'Iteration[{self.scenario_id}][{", ".join(settings)}]'

    def to_dict(self) -> Dict[str, Any]:
        return {
            "execution_id": self.execution_id,
            "scenario_id": self.scenario_id,
            "scenario_settings": self.scenario_settings,
        }

    @classmethod
    def from_dict(cls, iteration: Dict[str, Any]) -> "OptimizationIteration":
        return cls(
            iteration["scenario_id"],
            iteration["execution_id"],
            iteration["scenario_settings"],
        )


class BaseOptimizationOrchestrator(ABC):
    def __init__(self, client: NumerousClient, parallelism: int):
        self.client = client
        self.target_job_id = self.client.optimization_config.target_job_id
        self.target_scenario_id = self.client.optimization_config.target_scenario_id
        self.aggregator_config = self.client.optimization_config.aggregator_config
        self.goal_function_config = self.client.optimization_config.goal_function_config
        self.goal_function_inputs = self.client.optimization_config.goal_function_inputs
        self.settings = self.client.optimization_config.settings
        self.parallelism = parallelism
        self.clean_wait_time = 1
        self.previous_ongoing_iterations = 0
        self.previous_iteration_status: Dict[str, ScenarioStatus] = {}
        log.info("Initializing orchestrator with parallelism %s", self.parallelism)

        # load state
        self.aggregated: List[Tuple[OptimizationIteration, float]] = [
            (OptimizationIteration.from_dict(agg["iteration"]), agg["value"])
            for agg in client.state.get("aggregated", [])
        ]
        self.completed: List[OptimizationIteration] = [
            OptimizationIteration.from_dict(completed)
            for completed in client.state.get("completed", [])
        ]
        self.ongoing: List[OptimizationIteration] = [
            OptimizationIteration.from_dict(ongoing)
            for ongoing in client.state.get("ongoing", [])
        ]
        self.failed: List[OptimizationIteration] = [
            OptimizationIteration.from_dict(failed)
            for failed in client.state.get("failed", [])
        ]
        if self.aggregated or self.completed or self.ongoing or self.failed:
            log.info(
                "Loaded %s aggregated, %s completed, %s ongoing, and %s failed iterations",
                len(self.aggregated),
                len(self.completed),
                len(self.ongoing),
                len(self.failed),
            )

    def _aggregate(self):
        for iteration in list(
            self.completed
        ):  # Create a copy, to prevent deleting from iterated collection
            aggregate = self.aggregate(iteration)
            self.client.set_optimization_iteration_score(
                iteration.scenario_id, aggregate
            )
            log.info("Aggregated %s to %f", iteration.scenario_id, aggregate)
            self.aggregated.append((iteration, aggregate))
            self.completed.remove(iteration)
        self.client.state.set(
            "aggregated",
            [
                {"value": value, "iteration": iteration.to_dict()}
                for iteration, value in self.aggregated
            ],
        )

    def _clean(self):
        finished = []
        failed = []

        if len(self.ongoing) != self.previous_ongoing_iterations:
            log.info("Awaiting finished iterations (%s ongoing)", len(self.ongoing))
            self.previous_ongoing_iterations = len(self.ongoing)

        for iteration in self.ongoing:
            doc, _ = self.client.get_scenario_document(scenario=iteration.scenario_id)
            status = (
                doc.get("jobs", {})
                .get(self.target_job_id, {})
                .get("status", {})
                .get("status", {})
            )

            if status != self.previous_iteration_status.get(iteration.scenario_id):
                log.debug(
                    "Iteration %s now %s", iteration.scenario_id, ScenarioStatus(status)
                )
                self.previous_iteration_status[iteration.scenario_id] = status
            if status == ScenarioStatus.FINISHED:
                finished.append(iteration)
            if status == ScenarioStatus.FAILED:
                failed.append(iteration)

        for iteration in finished:
            self.ongoing.remove(iteration)
            self.completed.append(iteration)
        for iteration in failed:
            self.ongoing.remove(iteration)
            self.failed.append(iteration)

        self.client.state.set(
            "ongoing", [ongoing.to_dict() for ongoing in self.ongoing]
        )
        self.client.state.set("failed", [failed.to_dict() for failed in self.failed])
        self.client.state.set(
            "completed", [completed.to_dict() for completed in self.completed]
        )

    def _launcher(self):
        log.info("Launching %s iterations...", len(self.iteration_configs))
        while self.iteration_configs:
            while not self._can_launch():
                sleep(0.1)
            self._launch(self.iteration_configs.pop())
        log.info("Launcher finished")

    def _launch(self, settings: List[ScenarioSetting]):
        scenario_id, execution_id = self.client.start_iteration_scenario(
            self.target_scenario_id, self.target_job_id, settings
        )
        iteration = OptimizationIteration(scenario_id, execution_id, settings)
        self.ongoing.append(iteration)
        self.client.state.set(
            "ongoing", [ongoing.to_dict() for ongoing in self.ongoing]
        )
        log.info("Started %s", iteration)
        self._running_progress()

    def _can_launch(self):
        return len(self.ongoing) < self.parallelism

    def _running_progress(self):
        finished = len(self.aggregated) + len(self.completed) + len(self.failed)
        total = finished + len(self.iteration_configs) + len(self.ongoing)
        progress = 100 * float(finished) / float(total)
        self.client.set_scenario_progress(
            "Running iterations", ScenarioStatus.RUNNING, progress, force=True
        )

    def _goal_function_progress(self):
        self.client.set_scenario_progress(
            "Searching for optimal iteration", ScenarioStatus.RUNNING, force=True
        )

    @abstractmethod
    def aggregate(self, iteration: OptimizationIteration) -> float:
        pass

    @abstractmethod
    def get_iteration_configs(self) -> List[List[ScenarioSetting]]:
        pass

    @abstractmethod
    def goal_function(self, current: float, best: float) -> float:  # noqa: F841
        pass

    def run(self, poll_timeout: float = 1.0) -> Tuple[OptimizationIteration, Any]:
        self.client.set_scenario_progress(
            "Initializing optimization", ScenarioStatus.INITIALZING, force=True
        )
        iteration_configs = self.client.state.get("settings", None)
        self.iteration_configs = (
            self.get_iteration_configs()
            if iteration_configs is None
            else iteration_configs
        )
        self.client.state.set("settings", self.iteration_configs)
        self._running_progress()
        launcher_thread = threading.Thread(name="Launcher", target=self._launcher)
        launcher_thread.start()
        while self.ongoing or launcher_thread.is_alive():
            self._clean()
            if self.completed:
                self._aggregate()
            else:
                sleep(poll_timeout)

        log.info(
            "Iteration and aggregation done (%s aggregates, %s failed), finding optimal scenario...",
            len(self.aggregated),
            len(self.failed),
        )
        candidate_iteration = None
        candidate_aggregate = None
        for iteration, aggregate in self.aggregated:
            self._goal_function_progress()
            if (
                candidate_iteration is None
                or self.goal_function(aggregate, candidate_aggregate) == aggregate
            ):
                candidate_iteration = iteration
                candidate_aggregate = aggregate

        self.client.state.unset("aggregated")
        self.client.state.unset("ongoing")
        self.client.state.unset("completed")
        self.client.state.unset("failed")
        self.client.state.unset("settings")

        if self.failed:
            raise RuntimeError(
                f"Optimization errors! {len(self.failed)} iterations failed"
            )
        elif candidate_iteration is None or candidate_aggregate is None:
            raise RuntimeError("Optimization did not find an optimal solution")
        else:
            log.info(
                "Optimization finished: Iteration=%s, Aggregate=%f",
                candidate_iteration,
                candidate_aggregate,
            )
            self.client.set_optimization_result(
                candidate_iteration.scenario_id,
                candidate_aggregate,
                candidate_iteration.scenario_settings,
            )
            return candidate_iteration, candidate_aggregate



import threading
from datetime import datetime, timedelta
from time import sleep


class NumerousLogHandler:  # TODO: inherit from logging.Handler
    def __init__(self, client, admin_log_id=None, echo=False):
        self.client = client
        self.buffered_records = []
        self._buffer_lock = threading.Lock()
        self.last_push = datetime.utcnow() - timedelta(hours=1)
        self.debounce = 5
        self.admin_log_id = admin_log_id
        self.echo = echo

        self.stop_timed_flush = threading.Event()

        def timed_flush(stop_event: threading.Event):
            while not stop_event.is_set():
                now_ = datetime.utcnow()
                if (now_ - self.last_push).seconds > self.debounce:
                    self.last_push = now_
                    self.flush()
                sleep(1)

        threading.Thread(
            target=timed_flush, args=(self.stop_timed_flush,), daemon=True
        ).start()

    def handle(self, record):
        self.log(record.msg)

    def log(self, message):
        now_ = datetime.utcnow()
        record = (now_, message)
        with self._buffer_lock:
            self.buffered_records.append(record)
        if self.echo:
            print(record)

    def flush(self):
        to_push = None
        with self._buffer_lock:
            if len(self.buffered_records) > 0:
                to_push = self.buffered_records.copy()
                self.buffered_records = []
        if to_push:
            if self.admin_log_id is not None:
                self.client.push_scenario_logs_admin(to_push, exe_id=self.admin_log_id)
            else:
                self.client.push_scenario_logs(to_push)

    def close(self):
        self.stop_timed_flush.set()
        self.flush()

    # TODO make a handler thread to schedule update to server!



import sys
import threading
from datetime import datetime
from queue import Queue
from time import sleep, time
from typing import Any, Callable, Generator, Optional, Protocol

from numerous.client import config

from .common import log, trace

FLOAT_SIZE = sys.getsizeof(float(0))

WRITE_MAX_ATTEMPTS = 5
WRITE_ERROR_ATTEMPT_RESET_SECONDS = 10
WRITE_BACKOFF_START_DELAY = 1


class WriteMethodProtocol(Protocol):
    def __call__(
        self,
        data_generator: Generator[dict[str, float], None, None],
        scenario: Optional[str] = None,
        clear: bool = False,
        must_flush: Optional[Callable[[int, int], bool]] = None,
        execution: Optional[str] = None,
        last_item_index: int = 0,
    ) -> None:
        pass


class NumerousBufferedWriter:
    def __init__(
        self,
        write_method: WriteMethodProtocol,
        writer_closed_method,
        scenario: str,
        buffer_size: int = 24 * 7,
        logger: Optional[Any] = None,
        trace: bool = False,
        queue_max_size: int = 100,
    ):
        self.scenario = scenario
        self._tags = None
        self.ix = 0
        self._write_method = write_method
        self._writer_closed = writer_closed_method
        self._buffer = None
        self.buffer_size = buffer_size
        self._write_queue: Queue[dict[str, float]] = Queue(queue_max_size)
        self.closed = False

        self.logger = logger  # for trace decorator
        self.trace = trace  # for trace decorator

        self.last_flush = time()
        self.max_elapse_flush = 600
        self.min_elapse_flush = 1
        self.force_flush = False
        self.buffer_number_size = 1e6
        self._last_item_index = 0
        self._restore_row: Optional[dict[str, float]] = None
        self._last_row: Optional[dict[str, float]] = None
        self._writer_thread = threading.Thread(
            target=self._writer_thread_func, daemon=True
        )
        self._writer_thread.start()

    def _write_generator(self) -> Generator[dict[str, float], None, None]:
        while True:
            if row := self._restore_row:
                self._restore_row = None
            else:
                row = self._write_queue.get()
                if row == "STOP":
                    return
            self._last_row = row
            self._last_item_index += 1
            yield row

    def _writer_thread_func(self) -> None:
        last_error_time: Optional[float] = None
        failed_attempts: int = 0

        while failed_attempts < WRITE_MAX_ATTEMPTS:
            try:
                log.debug("Starting writer, attempt=%s", failed_attempts)
                self._write_method(
                    self._write_generator(),
                    must_flush=self._must_flush,
                    last_item_index=self._last_item_index,
                )
                return
            except Exception:
                elapsed_since_last_error = (
                    time() - last_error_time if last_error_time else 0.0
                )
                if elapsed_since_last_error > WRITE_ERROR_ATTEMPT_RESET_SECONDS:
                    failed_attempts = 0

                log.exception("An error occured writing data, retrying...")
                back_off_time = WRITE_BACKOFF_START_DELAY * 2**failed_attempts
                self._restore_row = self._last_row
                failed_attempts += 1
                sleep(back_off_time)
        log.error("Shutting down writer thread due to many attempts")

    def _must_flush(self, n_rows, rows_size):
        estimated_size = n_rows * rows_size * FLOAT_SIZE
        elapsed_since_last_flush = time() - self.last_flush
        if self.force_flush:
            self.force_flush = False
            self.last_flush = time()
            return True
        elif (
            n_rows > self.buffer_size
            and elapsed_since_last_flush > self.min_elapse_flush
        ) or estimated_size * 1.1 > config.GRPC_MAX_MESSAGE_SIZE:
            self.last_flush = time()
            return True
        elif elapsed_since_last_flush > self.max_elapse_flush:
            self.last_flush = time()
            return True
        return False

    def _init_buffer(self):
        self._buffer_count = 0
        self._buffer_timestamp = []

    @trace
    def write_row(self, data):
        if "_index" not in data:
            data["_index"] = self.ix
            self.ix += 1

        if isinstance(data["_index"], datetime):
            data["_index"] = data["_index"].timestamp()
        if self.closed:
            raise ValueError("Queue is closed!")
        self._write_queue.put(data)
        return data["_index"]

    @trace
    def flush(self):
        self.force_flush = True
        self._init_buffer()

    @trace
    def close(self):
        if not self.closed:
            self.flush()
            self._write_queue.put("STOP")
            self._writer_thread.join()
            self._writer_closed(self.scenario)

        self.closed = True



import logging
import sys
import tarfile
from enum import Enum
from fnmatch import fnmatch
from hashlib import sha256
from io import BytesIO
from os import linesep
from pathlib import Path, PurePosixPath
from typing import Any, Callable, ContextManager, Generator, Iterable, Optional
from urllib.parse import urlparse
from uuid import uuid4

import grpc
from colorama import Fore, Style
from colorama import init as init_colorama
from grpc_interceptor import ClientCallDetails, ClientInterceptor
from numerous.grpc.spm_pb2_grpc import TokenManagerStub
from pathspec import PathSpec

log = logging.getLogger("__name__")


init_colorama()


def red(msg: Any) -> str:
    return f"{Fore.RED}{msg}{Fore.RESET}"


def green(msg: Any) -> str:
    return f"{Fore.GREEN}{msg}{Fore.RESET}"


def cyan(msg: Any) -> str:
    return f"{Fore.CYAN}{msg}{Fore.RESET}"


def yellow(*msg: Any, sep: str = " ") -> str:
    return f"{Fore.YELLOW}{sep.join(msg)}{Fore.RESET}"


def bold(msg: Any) -> str:
    return f"{Style.BRIGHT}{msg}{Style.NORMAL}"


def print_error_and_exit(msg: Any, exit_code: int = 1):
    print(red(msg))
    sys.exit(exit_code)


def print_repository_help_message(path: Path):
    print(
        f" ! {bold('Remember to change directory to the repository:')} {bold(cyan(f'cd {path}'))}"
    )
    print(f" * Connect a remote: {bold(cyan('numerous config --remote <REMOTE URL>'))}")
    print(
        f" * Checkout a scenario: {bold(cyan('numerous checkout --scenario <SCENARIO ID>'))}"
    )
    print(f" * Push changes: {bold(cyan('numerous push -c <COMMENT>'))}")
    print(f" * Build code and update scenario: {bold(cyan('numerous build'))}")


def get_grpc_channel(
    url: str,
    auth_plugin: Optional[grpc.AuthMetadataPlugin],
    extra_interceptors: Optional[list[ClientInterceptor]],
) -> grpc.Channel:
    parsed_url = urlparse(url)

    options = (
        ("grpc.keepalive_time_ms", 10000),
        ("grpc.keepalive_timeout_ms", 5000),
        ("grpc.keepalive_permit_without_calls", 1),
        ("grpc.http2_max_pings_without_data", 9999),
    )

    if parsed_url.scheme == "https":
        channel = grpc.secure_channel(
            f"{parsed_url.hostname}:{parsed_url.port or '443'}",
            grpc.composite_channel_credentials(
                grpc.ssl_channel_credentials(),
                grpc.metadata_call_credentials(auth_plugin),
            )
            if auth_plugin
            else grpc.ssl_channel_credentials(),
            options=options,
        )
    elif parsed_url.hostname == "localhost":
        log.debug("Getting local channel")
        channel = grpc.secure_channel(
            f"{parsed_url.hostname}:{parsed_url.port or '50056'}",
            grpc.local_channel_credentials(),
            options=options,
        )
    else:
        channel = grpc.insecure_channel(
            f"{parsed_url.hostname}:{parsed_url.port or '80'}", options=options
        )

    if extra_interceptors:
        channel = grpc.intercept_channel(channel, *extra_interceptors)

    return channel


class OrganizationInterceptor(ClientInterceptor):
    def __init__(self, organization_slug: str):
        self._organization_slug = organization_slug

    def intercept(
        self,
        method: Callable,
        request_or_iterator: Any,
        call_details: grpc.ClientCallDetails,
    ):
        new_metadata = (call_details.metadata or []) + [
            ("numerous-organization", self._organization_slug)
        ]

        new_details = ClientCallDetails(
            method=call_details.method,
            timeout=call_details.timeout,
            metadata=new_metadata,
            credentials=call_details.credentials,
            wait_for_ready=call_details.wait_for_ready,
            compression=call_details.compression,
        )

        return method(request_or_iterator, new_details)


class StubClient:
    def __init__(
        self,
        api_url: str,
        stub: Callable[[grpc.Channel], Any],
        plugin: Optional[grpc.AuthMetadataPlugin] = None,
        extra_interceptors: Optional[list[ClientInterceptor]] = None,
    ):
        self.channel = get_grpc_channel(api_url, plugin, extra_interceptors)
        self.stub = stub(self.channel)

    def __enter__(self):
        return self.stub

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.channel.close()


def get_token_manager(
    api_url: str, organization_slug: str
) -> ContextManager[TokenManagerStub]:
    return StubClient(
        api_url,
        TokenManagerStub,
        extra_interceptors=[OrganizationInterceptor(organization_slug)],
    )


class GzippedTarball:
    def __init__(self, source: Path, paths: Iterable[Path]):
        self.source = source
        self.files = list(path for path in paths if path.is_file())
        self.tar_path = source / f"{uuid4()}.tar.gz"

    def __enter__(self):
        with tarfile.open(str(self.tar_path), mode="w:gz") as tar:
            for file_path in self.files:
                info = tarfile.TarInfo(str(PurePosixPath(Path(file_path.name))))
                info.path = str(file_path.relative_to(self.source))
                info.size = file_path.stat().st_size
                with file_path.open("rb") as file:
                    tar.addfile(info, fileobj=file)
        return self.tar_path

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.tar_path.unlink()


def display_error_and_exit(
    error_header: str, help_lines: list[str], exit_code: int = 1
):
    print(red(error_header))
    print(*help_lines, sep=linesep)
    sys.exit(exit_code)


def extract_gzipped_tarball(file: BytesIO, path: Path) -> None:
    with tarfile.open(fileobj=file, mode="r:gz") as tar:
        tar.extractall(str(path))


def exclude_paths(
    root: Path,
    paths: Iterable[Path],
    exclude_patterns: list[str],
) -> Generator[Path, None, None]:
    spec = PathSpec.from_lines("gitwildmatch", exclude_patterns)
    for path in paths:
        relative_path = path.relative_to(root)
        exclude = spec.match_file(str(relative_path))
        if not exclude:
            yield path


def walk(directory: Path, include_pattern: Optional[str] = None):
    yield from _walk(directory, directory, include_pattern)


def _walk(
    root: Path, directory: Path, include_pattern: Optional[str] = None
) -> Generator[Path, None, None]:
    for path in directory.iterdir():
        relative_path = path.relative_to(root)
        if include_pattern and not fnmatch(str(relative_path), include_pattern):
            continue

        yield path
        if path.is_dir():
            yield from _walk(root, path, include_pattern)


def hash_directory(directory: Path) -> str:
    dirhash = sha256()
    for file_path in (path for path in walk(directory) if path.is_file()):
        dirhash.update(str(file_path.lstat().st_mtime).encode("utf-8"))
    return dirhash.hexdigest()


class RepositoryConfigLoadWarning(Enum):
    REMOTE_INVALID = 0
    PROJECT_SCENARIO_INVALID = 1
    PROJECT_SCENARIO_MISSING = 2
    COMMIT_MISSING = 3
    SNAPSHOT_MISSING = 4


def print_load_warnings(warnings: list[RepositoryConfigLoadWarning]) -> None:
    if warnings:
        print(
            yellow(
                f"{len(warnings)} where found while loading repository configuration"
            )
        )

    for warning in warnings:
        if warning == RepositoryConfigLoadWarning.COMMIT_MISSING:
            print(
                yellow(
                    bold("Commit missing:"), "Current commit will not be highlighted."
                )
            )
        elif warning == RepositoryConfigLoadWarning.PROJECT_SCENARIO_INVALID:
            print(yellow(bold("Invalid scenario:"), "Checkout needed."))
        elif warning == RepositoryConfigLoadWarning.PROJECT_SCENARIO_MISSING:
            print(yellow(bold("No scenario:"), "Checkout needed"))
        elif warning == RepositoryConfigLoadWarning.REMOTE_INVALID:
            print(
                yellow(
                    bold("Invalid remote:"),
                    "The configured remote is invalid, you must run "
                    "'numerous config --remote REMOTE' again",
                )
            )
        elif warnings == RepositoryConfigLoadWarning.SNAPSHOT_MISSING:
            print(
                yellow(
                    bold("Snapshot missing:"),
                    "No snapshot is defined, you must push with "
                    "'numerous push -c COMMENT'",
                )
            )


class NumerousCommandLineError(Exception):
    pass



import time

from numerous.grpc.spm_pb2 import RefreshRequest, Token
from numerous.grpc.spm_pb2_grpc import TokenManagerStub


class RefreshingAccessToken:
    _refresh_threshold = 8 * 60

    def __init__(self, token_manager: TokenManagerStub, refresh_token: str):
        self._token_manager = token_manager
        self._refresh_token = refresh_token
        self._access_token: str = ""
        self._last_refresh_timestamp = float("-inf")

    @property
    def value(self) -> str:
        self._refresh_if_need()
        return self._access_token

    def _refresh_if_need(self):
        now = time.monotonic()
        if now - self._last_refresh_timestamp >= self._refresh_threshold:
            self._access_token = self._token_manager.GetAccessToken(
                RefreshRequest(refresh_token=Token(val=self._refresh_token))
            ).val
            self._last_refresh_timestamp = now

    def __eq__(self, other):
        if isinstance(other, RefreshingAccessToken):
            return other._access_token == self._access_token
        else:
            return False



import enum
import json
from dataclasses import dataclass
from typing import Any, Union


@dataclass
class Parameter:  # will be used as a base-class later for implementing specific functionality
    id: str
    uuid: str
    display_name: str
    type: str
    value: Any

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Parameter":
        return Parameter(
            id=data["id"],
            uuid=data["uuid"],
            display_name=data["displayName"],
            type=data["type"],
            value=parse_value(data["value"], data["type"]),
        )


@dataclass
class FileReference:
    id: str
    path: str
    name: str

    @staticmethod
    def from_document(data: dict[str, Any]) -> "FileReference":
        return FileReference(data["id"], data["path"], data["name"])


class SimTimeUnit(enum.Enum):
    years = "years"
    months = "months"
    days = "days"
    hours = "hours"
    minutes = "minutes"
    seconds = "seconds"

    @classmethod
    def from_string(cls, value_str) -> "SimTimeUnit":
        for member in cls:
            if member.value == value_str:
                return member
        raise ValueError(f"No matching enum member for value: {value_str}")


@dataclass
class Currency:
    name: str
    code: str

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Currency":
        return Currency(data["name"], data["code"])


@dataclass
class Month:
    value: int

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Month":
        return Month(data["value"])


@dataclass
class MapLocation:
    id: str
    name: str
    center: tuple[float, float]

    @staticmethod
    def from_document(data: dict[str, Any]) -> "MapLocation":
        return MapLocation(data["id"], data["name"], data["center"])


@dataclass
class TimeValue:
    value: float
    unit: SimTimeUnit

    @staticmethod
    def from_document(data: dict[str, Any]) -> "TimeValue":
        return TimeValue(float(data["value"]), SimTimeUnit.from_string(data["unit"]))


@dataclass
class Config:
    value: str

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Config":
        return Config(data["value"])


@dataclass
class Selector:
    value: Union[str, float, int]

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Selector":
        return Selector(data["value"])


def parse_value(value: Any, type_: Any) -> Any:
    if type_ == "string":
        return value
    elif type_ == "number":
        return float(value)
    elif type_ == "boolean":
        return value
    elif type_ == "file":
        return FileReference.from_document(value)
    elif type_ == "json":
        return json.loads(value)
    elif type_ == "time_value":
        return TimeValue.from_document(value)
    elif type_ == "currency":
        return Currency.from_document(value)
    elif type_ == "month":
        return Month.from_document(value)
    elif type_ == "map":
        return MapLocation.from_document(value)
    elif type_ == "config":
        return Config.from_document(value)
    elif type_ == "selecter":
        return Selector.from_document(value)
    raise ValueError(f"Unkwnown parameter type: {type_}")


def get_parameters_dict(data: list[dict[str, Any]]) -> dict[str, Parameter]:
    return {param.id: param for param in get_parameters_list(data)}


def get_parameters_list(data: list[dict[str, Any]]) -> list[Parameter]:
    return [Parameter.from_document(param) for param in data]



import uuid
from typing import Any, Callable, Optional

import grpc
from grpc_interceptor import ClientCallDetails, ClientInterceptor


class ValidationInterceptor(ClientInterceptor):
    def __init__(
        self, token, token_callback: Callable[[], str], instance: Optional[str] = None
    ):
        self.token = token
        self.instance = str(uuid.uuid4()) if instance is None else instance
        self.token_callback = token_callback

    def intercept(
        self,
        method: Callable,
        request_or_iterator: Any,
        call_details: grpc.ClientCallDetails,
    ):
        new_details = ClientCallDetails(
            call_details.method,
            call_details.timeout,
            [("token", self.token_callback()), ("instance", str(self.instance))],
            call_details.credentials,
            call_details.wait_for_ready,
            call_details.compression,
        )

        return method(request_or_iterator, new_details)



import json
from pathlib import Path, PurePosixPath
from typing import Generator, Optional
from urllib.parse import urlparse

from numerous.cli.utils import (
    NumerousCommandLineError,
    RepositoryConfigLoadWarning,
    exclude_paths,
    hash_directory,
    walk,
)


class InvalidRepositoryRemoteUrl(NumerousCommandLineError):
    def __init__(self, remote_url: str):
        self.remote_url = remote_url

    def __str__(self) -> str:
        return f'Invalid repository remote URL "{self.remote_url}"'


class RepositoryRemote:
    def __init__(
        self,
        url: Optional[str] = None,
        api_url: Optional[str] = None,
        organization: Optional[str] = None,
        job_id: Optional[str] = None,
        name: Optional[str] = None,
    ):
        if name and api_url and job_id and organization:
            self.name = name
            self.api_url = api_url
            self.job_id = job_id
            self.organization = organization
        elif url is not None:
            parsed_url = urlparse(url)
            try:
                _, self.organization, self.job_id, self.name = PurePosixPath(
                    parsed_url.path
                ).parts
            except ValueError as error:
                raise InvalidRepositoryRemoteUrl(url) from error
            self.api_url = f"{parsed_url.scheme}://{parsed_url.netloc}"
        else:
            raise ValueError(
                "RepositoryRemote must be supplied 'url' or 'api_url', 'job_id' and 'name'."
            )

    def __str__(self):
        return f"{self.api_url}/{self.organization}/{self.job_id}/{self.name}"


class ProjectScenario:
    def __init__(
        self,
        project_scenario: Optional[str] = None,
        project_id: Optional[str] = None,
        scenario_id: Optional[str] = None,
    ):
        if project_scenario:
            self.project_id, self.id = project_scenario.split(":")
        elif project_id is not None and scenario_id is not None:
            self.project_id = project_id
            self.id = scenario_id
        else:
            raise ValueError(
                "ProjectScenario must be supplied 'project_scenario' or 'project_id' and 'scenario_id"
            )

    def __str__(self):
        return f"{self.project_id}:{self.id}"


class NumerousRepository:
    FILE_NAME = ".numerous.repository.json"

    def __init__(self, path: Optional[Path] = None):
        self.path: Path = path or Path.cwd()
        self.refresh_token: Optional[str] = None
        self.remote: Optional[RepositoryRemote] = None
        self.scenario: Optional[ProjectScenario] = None
        self.commit: Optional[str] = None
        self.snapshot: Optional[str] = None

    def save(self) -> "NumerousRepository":
        with (self.path / NumerousRepository.FILE_NAME).open("w") as fp:
            json.dump(
                {
                    "scenario": None if self.scenario is None else str(self.scenario),
                    "commit": self.commit,
                    "snapshot": self.snapshot,
                    "refresh_token": self.refresh_token,
                    "remote": None if self.remote is None else str(self.remote),
                },
                fp,
            )
            return self

    def load(self) -> tuple["NumerousRepository", list[RepositoryConfigLoadWarning]]:
        with (self.path / NumerousRepository.FILE_NAME).open("r") as fp:
            config = json.load(fp)
            return self, list(self._parse_config(config))

    def _parse_config(
        self, config
    ) -> Generator[RepositoryConfigLoadWarning, None, None]:
        try:
            self.remote = (
                None
                if config.get("remote") is None
                else RepositoryRemote(url=config["remote"])
            )
        except ValueError:
            yield RepositoryConfigLoadWarning.REMOTE_INVALID

        if "scenario" not in config:
            yield RepositoryConfigLoadWarning.PROJECT_SCENARIO_MISSING

        try:
            self.scenario = (
                None
                if config.get("scenario") is None
                else ProjectScenario(config["scenario"])
            )
        except ValueError:
            yield RepositoryConfigLoadWarning.PROJECT_SCENARIO_INVALID

        if "commit" not in config:
            yield RepositoryConfigLoadWarning.COMMIT_MISSING
        self.commit = config.get("commit")

        if "snapshot" not in config:
            yield RepositoryConfigLoadWarning.SNAPSHOT_MISSING

        self.snapshot = config.get("snapshot")
        self.refresh_token = config.get("refresh_token")

    def is_dirty(self) -> bool:
        return hash_directory(self.path) != self.snapshot

    def walk(self):
        exclude_file = self.path / ".exclude"
        exclude_patterns = (
            exclude_file.open("r").readlines() if exclude_file.exists() else []
        )
        yield from exclude_paths(self.path, walk(self.path), exclude_patterns)



"""numerous.sdk.connect - connect your code to the numerous platform!

The connect package contains all the functionality needed to build applications
that connect to and take advantage of the numerous platform.
"""

from .job_client import JobClient, JobStatus

__all__ = ("JobClient", "JobStatus")



import logging

from numerous.image_tools.app import run_job
from numerous.image_tools.job import NumerousSimulationJob

logger = logging.getLogger("multiply-by-two")

#  Creates the MultiplyByTwoJob inheriting the NumerousSimulationJob, which uses the standard work-flow of reading data,
#  stepping the model by calculating new states, outputs, and the next time step, and saves these outputs to numerous.


class MultiplyByTwoJob(NumerousSimulationJob):
    def __init__(self):
        super(MultiplyByTwoJob, self).__init__()

    def initialize_simulation_system(self):
        self.numerous_name = "example"
        # Set align_outputs_to_next_timestep to False, when outputs are to be aligned to the current timestep in the
        # step() method
        self.align_outputs_to_next_timestep = False

    """
    The step method is called each time the inputs have been read, and must return the next timestep and the outputs,
    which is a dictionary formatted as {tag1: value1, tag2: value2...}
    """

    def step(self, t: float, dt: float) -> tuple[float, dict]:
        component = self.system.components[  # type: ignore[union-attr]
            self.numerous_name
        ]  # Get the component named 'example' from the frontend
        inputs = (
            component.inputs
        )  # This dict contains the inputs to the named component
        model = (
            component.model
        )  # This allows the user to access the user instantiated python model class
        outputs = model.multiply_by_two(inputs, t)  # a dict with outputs
        # we can add more outputs.. for example output resumed True if job is resumed
        outputs.update({"resumed": len(self.system.states) > 0})  # type: ignore[union-attr]
        # Finally return the next timestep (t+dt) and the outputs. The outputs are timestamped with the current
        # timestep (t), because of the 'align_outputs_to_next_timestep = False' set above.
        return t + dt, outputs

    # you can save states, by defining the serialize_states. This way you can resume a stateful model.
    def serialize_states(self, t: float) -> bool:
        return True


def run_example():
    run_job(
        numerous_job=MultiplyByTwoJob(),
        appname="multiply-by-two",
        model_folder="numerous.image_tools.example_models",
    )


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    run_example()



import json
from typing import Any, List

from numerous.grpc import spm_pb2
from numerous.grpc.spm_pb2_grpc import SPMStub


def load_target_scenario_data_and_set_job_id(
    spm_stub: SPMStub, project_id: str, scenario: dict[str, Any]
) -> List[Any]:
    target_scenario_id = scenario["optimizationTargetScenarioID"]
    target_scenario_request = spm_pb2.Scenario(
        project=project_id, scenario=target_scenario_id
    )
    target_scenario_response = spm_stub.GetScenario(target_scenario_request)
    target_scenario = json.loads(target_scenario_response.scenario_document)

    if not isinstance(target_scenario, dict):
        raise TypeError(f"Received invalid target scenario data: {target_scenario}")

    target_job_id = next(
        job_id
        for job_id, job in target_scenario["jobs"].items()
        if job.get("isMain", False)
    )
    return [target_scenario, target_job_id]



import sys
from pathlib import Path
from typing import Optional

from ..repository import NumerousRepository
from ..utils import bold, green, print_load_warnings, print_repository_help_message, red


def command_init(path: Optional[Path]):
    path = path or Path.cwd()

    if (path / NumerousRepository.FILE_NAME).exists():
        try:
            repo, load_warnings = NumerousRepository(path).load()
            print_load_warnings(load_warnings)
            if repo.remote:
                error = f"Folder already a repository for {bold(repo.remote.name)}"
            else:
                error = "Folder already a repository without a configured remote"
            print(red(f"Cannot init: {error}"))
        except Exception:
            print(
                red(
                    "Cannot init: Folder already exists, but repository configuration is invalid"
                )
            )
        finally:
            sys.exit(1)

    path.mkdir(parents=True, exist_ok=True)

    if not (path / ".exclude").exists():
        src = Path(__file__).parent.parent / ".exclude"
        dst = path / ".exclude"
        dst.write_text(src.read_text())

    repo = NumerousRepository(path).save()

    abs_path = Path(path.name).absolute()
    print(green(f"Initialized new repository in {bold(abs_path)}"))
    print_repository_help_message(abs_path)



import json
from datetime import datetime, timedelta
from enum import Enum
from time import sleep
from typing import Iterable, Iterator, Optional, Sequence, TypeVar

import grpc
from numerous.grpc import spm_pb2, spm_pb2_grpc

from numerous.sdk.connect.hibernation import HibernationHandler
from numerous.sdk.connect.job_utils import JobIdentifier
from numerous.sdk.models.input import (
    InputSource,
    InputVariable,
    InputVariableScenarioSource,
    InputVariableStaticSource,
)
from numerous.sdk.models.job import Job


class HibernatingSourceAction(Enum):
    HIBERNATE_SELF = 0
    WAIT = 1


INPUT_READER_DEFAULT_BUFFER_SIZE = 10000
HIBERNATE_SELF = HibernatingSourceAction.HIBERNATE_SELF
nan = float("nan")


class ReadRetryExhausted(Exception):
    """Raised when reading has been retried too many times."""


class InputSourceHibernating(Exception):
    """Raised when reaching the end of a hibernating input source."""

    def __init__(self, stream: "InputStream", key: datetime) -> None:
        self.stream = stream
        self.key = key


class DataListWrapper:
    """A helper class that wraps and chains data lists."""

    def __init__(
        self,
        data_lists: list[spm_pb2.DataList],
        repeat_offset: float,
        previous: Optional["DataListWrapper"] = None,
    ) -> None:
        self._index = next(
            (
                list(block.values)
                for data_list in data_lists
                for block in data_list.data
                if block.tag == "_index"
            )
        )
        self._repeat_offset = repeat_offset
        self._previous = previous
        self.min_key = min(self._index)
        self.max_key = max(self._index)
        self.last_key_step = self._find_last_key_step()

        self._values = {
            block.tag: list(block.values)
            for data_list in data_lists
            for block in data_list.data
            if block.tag != "_index"
        }
        self._reference_count = 0

    def _find_last_key_step(self) -> Optional[float]:
        if not self._index:
            if self._previous is not None:
                return self._previous.last_key_step
            else:
                return None

        if len(self._index) > 1:
            return self._index[-1] - self._index[-2]

        if self._previous is None:
            return None
        if last_index := self._previous.last_index():
            return self._index[-1] - last_index
        else:
            return None

    def last_index(self) -> Optional[float]:
        if self._index:
            return self._index[-1]
        elif self._previous:
            return self.last_index()
        else:
            return None

    def remove_previous(self) -> None:
        self._previous = None

    def last_value(self, tag: str) -> float:
        return self._values[tag][-1]

    def is_after(self, key: float) -> bool:
        upper_boundary = self.max_key + self._repeat_offset
        return key > upper_boundary

    def is_before(self, key: float) -> bool:
        lower_boundary = self.min_key + self._repeat_offset
        return key < lower_boundary

    def read_tag(self, tag: str, key: float) -> float:
        offset_cancelled_key = key - self._repeat_offset
        previous_value: Optional[float] = None
        for value_key, value in zip(self._index, self._values[tag]):
            if offset_cancelled_key >= value_key:
                previous_value = value
                continue
            elif previous_value is None:
                raise KeyError(
                    f"Key {key!r} outside lower bound of data list, and no previous "
                    "value."
                )
            else:
                return previous_value
        if previous_value is not None:
            return previous_value

        raise RuntimeError("Could not read tag with no data.")


class InputStream:
    """A helper class that reads data from a specific input source."""

    def __init__(
        self,
        spm_stub: spm_pb2_grpc.SPMStub,
        job_start_time: datetime,
        job_end_time: Optional[datetime],
        input_source: InputSource,
        input_variables: Iterable[InputVariable],
        repeat_spacing: Optional[float] = None,
    ) -> None:
        self._spm_stub = spm_stub
        self._job_start_time = job_start_time
        self._job_end_time = job_end_time
        self._input_source = input_source
        self._input_variables = self._filter_input_variables(input_variables)
        self._maximum_input_variable_offset = max(
            input_variable.offset for input_variable in self._input_variables
        )
        self._repeat_spacing = repeat_spacing
        self._repeat_start_index: Optional[float] = None
        self._repeat_offset = 0.0
        self._stream: Optional[Iterator[spm_pb2.DataList]] = None
        self._data_lists: list[DataListWrapper] = []
        self._hibernate_after_base_key: Optional[float] = None

    def _filter_input_variables(
        self,
        input_variables: Iterable[InputVariable],
    ) -> list[InputVariable]:
        return [
            input_variable
            for input_variable in input_variables
            if isinstance(input_variable.source, InputVariableScenarioSource)
            and input_variable.source.source == self._input_source
        ]

    def read_variables(self, key: datetime) -> dict[str, float]:
        return {var.key: self._read_variable(var, key) for var in self._input_variables}

    def _read_variable(self, variable: InputVariable, key: datetime) -> float:
        source = variable.source
        if not isinstance(source, InputVariableScenarioSource):
            raise ValueError()

        self._try_raise_source_hibernating(key)
        base_key = variable.job_time_to_base_time(key)
        previous_data_list: Optional[DataListWrapper] = None
        out_of_scope_data_lists: list[DataListWrapper] = []
        for data_list in self._data_lists:
            if data_list.is_after(base_key):
                self._collect_out_of_scope_data_list(
                    variable, out_of_scope_data_lists, previous_data_list, data_list
                )
                previous_data_list = data_list
            elif data_list.is_before(base_key):
                if not previous_data_list:
                    return nan
                else:
                    previous_value = previous_data_list.last_value(source.tag)
                    source.value = previous_value
                    return previous_value
            else:
                current_value = data_list.read_tag(source.tag, base_key)
                source.value = current_value
                return current_value

        self._data_lists.append(self._get_next_data_list(key))
        for out_of_scope_data_list in out_of_scope_data_lists:
            self._data_lists.remove(out_of_scope_data_list)
        return self._read_variable(variable, key)

    def _collect_out_of_scope_data_list(
        self,
        variable: InputVariable,
        out_of_scope_data_lists: list[DataListWrapper],
        old_previous_data_list: Optional[DataListWrapper],
        new_previous_data_list: DataListWrapper,
    ):
        if variable.offset != self._maximum_input_variable_offset:
            return

        if old_previous_data_list is None:
            return

        out_of_scope_data_lists.append(old_previous_data_list)
        new_previous_data_list.remove_previous()

    def _get_next_data_list(self, key: datetime) -> DataListWrapper:
        data_lists = self._get_next_complete_data_lists(key)
        if data_lists:
            previous = self._data_lists[-1] if self._data_lists else None
            next_data_list = DataListWrapper(
                data_lists,
                self._repeat_offset,
                previous=previous,
            )
            if self._repeat_start_index is None:
                self._repeat_start_index = next_data_list.min_key
            return next_data_list
        else:
            self._try_raise_source_hibernating(key)
            self._update_repeat_offset()
            self._stream = self._create_stream(key)
            next_data_list = self._get_next_data_list(key)
            self._repeat_start_index = next_data_list.min_key
            return next_data_list

    def _try_raise_source_hibernating(self, key: datetime) -> None:
        if self._hibernate_after_base_key is None:
            return

        if (
            self._input_source.job_time_to_base_time(key)
            > self._hibernate_after_base_key
        ):
            raise InputSourceHibernating(self, key)

    def _update_repeat_offset(self):
        if not self._data_lists or self._hibernate_after_base_key is not None:
            return

        last_data_list = self._data_lists[-1]
        self._repeat_offset += last_data_list.max_key
        self._repeat_offset -= self._repeat_start_index or 0.0

        if self._repeat_spacing is not None:
            self._repeat_offset += self._repeat_spacing
        elif last_data_list.last_key_step is not None:
            self._repeat_offset += last_data_list.last_key_step
        else:
            self._repeat_offset = 1.0

    def _get_next_complete_data_lists(self, key: datetime) -> list[spm_pb2.DataList]:
        if self._stream is None:
            self._stream = self._create_stream(key)

        data_lists: list[spm_pb2.DataList] = []
        while True:
            data_list = _next_grpc_retry(self._stream)
            if data_list and data_list.stream_status == "HIBERNATE":
                self._hibernate_after_base_key = self._get_hibernate_after_key(
                    data_lists
                )

            if data_list and data_list.data:
                data_lists.append(data_list)

            if data_list is None or data_list.block_complete or data_list.row_complete:
                break

        return data_lists

    def check_and_handle_resume_after_hibernation(self, key: datetime) -> bool:
        self._hibernate_after_base_key = None
        self._stream = self._create_stream(key)
        try:
            self.read_variables(key)
        except InputSourceHibernating:
            return False
        else:
            return True

    def _get_hibernate_after_key(
        self, data_lists: list[spm_pb2.DataList]
    ) -> Optional[float]:
        index = next(
            (
                block.values
                for data_list in data_lists
                for block in data_list.data
                if block.tag == "_index"
            ),
            None,
        )

        if index:
            return index[-1]
        elif self._data_lists:
            return self._data_lists[-1].last_index()
        else:
            return None

    def _create_stream(self, key: Optional[datetime]) -> Iterator[spm_pb2.DataList]:
        is_closed: bool = self._spm_stub.GetDataSetClosed(
            request=spm_pb2.Scenario(
                project=self._input_source.project_id,
                scenario=self._input_source.scenario_id,
            )
        ).is_closed
        maximum_variable_offset = max(
            variable.offset for variable in self._input_variables
        )
        tags = [
            input_variable.source.tag
            for input_variable in self._input_variables
            if isinstance(input_variable.source, InputVariableScenarioSource)
        ]

        if key is None:
            start_time = 0.0
        else:
            start_time = (
                key.timestamp()
                - self._input_source.offset
                - maximum_variable_offset
                - self._repeat_offset
            )

        if self._job_end_time is not None:
            end_time = self._job_end_time.timestamp() - self._input_source.offset
        else:
            end_time = float("inf")

        request = spm_pb2.ReadScenario(
            scenario=self._input_source.scenario_id,
            project=self._input_source.project_id,
            tags=tags,
            start=start_time,
            end=end_time,
            time_range=True,
            listen=not is_closed,
        )

        return self._read_data_with_retry(request)

    def _read_data_with_retry(
        self, request: spm_pb2.ReadScenario, max_attempts: int = 5
    ) -> Iterator[spm_pb2.DataList]:
        for attempt in range(max_attempts):
            try:
                return iter(self._spm_stub.ReadData(request=request))
            except grpc.RpcError:
                sleep(2**attempt)
        raise ReadRetryExhausted()


class InputReader(Iterable[tuple[datetime, dict[str, float]]]):
    """Reads input data according to the time setup of the current job.

    The input reader is meant to be used to iterate over the input data
    of the current job, taking into account the alignment of input sources,
    and input variables.

    During iteration the `value` attribute of :class:`InputVariable` objects in the
    scenario will be updated.
    """

    _streams: Sequence[InputStream]

    def __init__(
        self,
        spm_stub: spm_pb2_grpc.SPMStub,
        hibernation_handler: HibernationHandler,
        identity: JobIdentifier,
        execution_id: str,
        job: Job,
        input_sources: list[InputSource],
        input_variables: list[InputVariable],
        hibernating_source_action: HibernatingSourceAction,
        step: float = 1.0,
        repeat_spacing: Optional[float] = None,
    ) -> None:
        self._spm_stub = spm_stub
        self._hibernation_handler = hibernation_handler
        self._identity = identity
        self._execution_id = execution_id
        self._input_sources = input_sources
        self._input_variables = input_variables
        self._static_variable_data = {
            variable.key: variable.source.value
            for variable in input_variables
            if isinstance(variable.source, InputVariableStaticSource)
        }
        self._start = job.time.start
        self._end = job.time.end
        loaded_last_read_key = self.load_last_read_key()
        self._next_key = (
            self._start if loaded_last_read_key is None else loaded_last_read_key
        )
        self._step = step
        self._repeat_spacing = repeat_spacing
        self._hibernating_source_action = hibernating_source_action
        self._last_read_key: Optional[datetime] = None
        self._streams = self._open_streams()

    def _open_streams(self) -> list[InputStream]:
        return [
            InputStream(
                self._spm_stub,
                self._start,
                self._end,
                input_source,
                self._input_variables,
                self._repeat_spacing,
            )
            for input_source in self._input_sources
        ]

    def __iter__(self) -> "InputReader":
        return self

    def __next__(self) -> tuple[datetime, dict[str, float]]:
        this_key = self._next_key
        if self._end is not None and this_key > self._end:
            raise StopIteration()

        self._next_key += timedelta(seconds=self._step)
        row_data: Optional[dict[str, float]] = None

        while row_data is None:
            try:
                row_data = {
                    variable_key: value
                    for stream in self._streams
                    for variable_key, value in stream.read_variables(this_key).items()
                }
            except InputSourceHibernating as hibernating:
                self._handle_input_source_hibernating(hibernating)
        row_data.update(self._static_variable_data)

        self._last_read_key = this_key
        return (this_key, row_data)

    def _handle_input_source_hibernating(self, hibernating: InputSourceHibernating):
        if self._hibernating_source_action == HibernatingSourceAction.HIBERNATE_SELF:
            self.store_last_read_key()
            self._hibernation_handler.hibernate()
        else:
            resumed = False
            while not resumed:
                sleep(60.0)
                resumed = hibernating.stream.check_and_handle_resume_after_hibernation(
                    hibernating.key
                )

    def store_last_read_key(self) -> None:
        if self._last_read_key is None:
            return

        self._spm_stub.SetScenarioCustomMetaData(
            request=spm_pb2.ScenarioCustomMetaData(
                project=self._identity.project_id,
                scenario=self._identity.scenario_id,
                execution=self._execution_id,
                key=f"{self._identity.job_id}.hibernated_at_job_read_time",
                meta=json.dumps(self._last_read_key.isoformat()),
            )
        )

    def load_last_read_key(self) -> Optional[datetime]:
        hibernated_at_job_read_time = self._spm_stub.GetScenarioCustomMetaData(
            request=spm_pb2.ScenarioCustomMetaData(
                project=self._identity.project_id,
                scenario=self._identity.scenario_id,
                execution=self._execution_id,
                key=f"{self._identity.job_id}.hibernated_at_job_read_time",
            )
        ).meta
        if not hibernated_at_job_read_time:
            return None
        else:
            try:
                return datetime.fromisoformat(json.loads(hibernated_at_job_read_time))
            except (ValueError, TypeError, json.decoder.JSONDecodeError):
                return None


T = TypeVar("T")


def _next_grpc_retry(iterator: Iterator[T], max_attempts: int = 5) -> Optional[T]:
    for attempt in range(max_attempts):
        try:
            return next(iterator, None)
        except grpc.RpcError:
            sleep(2**attempt)
    raise ReadRetryExhausted()



import copy
import importlib
import logging
import os
import sys
from pathlib import Path
from time import sleep
from typing import Any, Iterable, Optional

import networkx as nx

from numerous.client import NumerousClient, ScenarioStatus
from numerous.client.data_source import DataSourceStreamStarting
from numerous.image_tools.errors import tb_str

logger = logging.getLogger("sim-setup")


class Component:
    def __init__(
        self,
        component_id: str,
        container_id: str,
        name: str,
        folder: str,
        path: str,
        handle: str,
        constants: dict,
        parameters: dict,
        import_flag: bool = False,
        item_class: Optional[str] = None,
        parent: Optional["Component"] = None,
    ):
        self._folder = Path(folder) if folder else None
        self._path = Path(path) if path else None
        self._handle = handle
        self.constants = constants
        self.parameters = parameters
        self._constructor = None
        self._entrypoint = None
        self.item = None
        self.item_class = item_class
        self.name = name
        self.component_id = component_id
        self.container_id = container_id
        self._import = import_flag
        self.inputs: dict[str, dict] = {}
        self.model: Any = None
        self.components: dict[str, Component] = {}
        self.parent: Optional[Component] = parent

    def _initialize(self, system, wrapper=None):
        if not self._import:
            return

        if wrapper:
            item = wrapper(self, self.name, system)
        else:
            item = self._entrypoint(self.name, system)

        self.item = item
        return item

    def _dynamicimport(self):
        if not self._import:
            return
        _entrypoint = getattr(
            importlib.import_module(self._path.as_posix()), self._handle
        )
        self._entrypoint = _entrypoint

    def _add_subcomponent(self, component):
        self.components.update({component.name: component})


class Project:
    def __init__(self, project: dict):
        self._project = project
        self.parameters = self._format_parameters()
        self.description = project.get("description")
        self._id = project.get("id")

    def _format_parameters(self):
        project_parameters = self._project.get("parameters")
        if not project_parameters:
            return {}

        return {
            project_parameter["id"]: project_parameter["value"]
            for project_parameter in project_parameters
        }


def unroll_components(
    components: Iterable[Component],
    graph: nx.DiGraph,
    references: list[str],
    container_references: dict[str, str],
) -> list[Component]:
    unrolled_components = []
    for component in components:
        unrolled_components.extend(
            unroll_components(
                component.components.values(),
                graph,
                references,
                container_references,
            )
        )

        graph.add_node(component.component_id, component=component)
        references.append(component.component_id)
        if component.container_id:
            container_references[component.container_id] = component.component_id

        unrolled_components.append(component)
    return unrolled_components


def sort_components(components: list[Component]) -> list[Component]:
    graph = nx.DiGraph()
    references: list[str] = []
    container_references: dict[str, str] = {}
    unrolled_components: list[Component] = unroll_components(
        components, graph, references, container_references
    )

    for component in unrolled_components:
        for param_name, param in component.constants.items():
            if isinstance(param, dict) and "ref" in param:
                ref = param["ref"]
                if ref in container_references:
                    ref = container_references[ref]
                    param["ref"] = ref
                if ref in references:
                    graph.add_edge(ref, component.component_id)
                else:
                    print(
                        f"Warning: {ref} is not defined. Referenced from {component.component_id}"
                    )

    components_ = nx.get_node_attributes(graph, "component")
    components_top_sorted = [components_[n] for n in nx.topological_sort(graph)]
    return components_top_sorted


class NumerousSystem:
    def __init__(
        self,
        nc: Optional[NumerousClient] = None,
        scenario=None,
        files=None,
        start_time=None,
        end_time=None,
        model_folder=None,
        numerous_job=None,
        states=None,
        t=None,
        dt=None,
        input=None,
        wrapper=None,
        load_model=True,
    ):
        self.references: dict[str, str] = {}
        self.nc = nc
        self.scenario = scenario
        self.files = files
        self.start_time = start_time
        self.end_time = end_time
        self._initialized = False
        self.logger = logging.getLogger("NumerousSystem")
        self._model_folder = model_folder
        self._has_model = False
        self.wrapper = wrapper
        self.numerous_job = numerous_job
        self.states = states
        self.dt = dt
        self.input = input

        # Initialize other system variables
        self.top_level_components = self._format_job_spec(self.scenario)
        self.unrolled_components = sort_components(self.top_level_components)
        for component in self.unrolled_components:
            if component._import:
                self._has_model = True
                break
        self.components = {
            component.name: component for component in self.unrolled_components
        }

        if load_model:
            self.load_model(t)

    def load_model(self, t):
        self.nc.set_scenario_progress(
            "Environment Initializing",
            ScenarioStatus.ENVIRONMENT_INITIALIZING,
            0.0,
            clean=True,
        )
        # Download needed files
        self._download_custom_files()
        self._dynamic_import_components()

        self.parameters = {}
        self.project = Project(project=self.nc.get_project_document())

        for par in self.scenario.get("parameters", []):
            if "id" in par and "value" in par:
                self.parameters.update({par.get("id"): par.get("value")})

        if self._has_model:
            while True:
                try:
                    data = self.input.get_at_time(t)
                    self.update_inputs(data)
                    self.instantiate_model()
                    self.nc.set_scenario_progress(
                        "model initialized", ScenarioStatus.INITIALZING, 0.0
                    )
                    break

                except DataSourceStreamStarting:
                    self.nc.set_scenario_progress(
                        "waiting for input stream to start",
                        ScenarioStatus.INITIALZING,
                        0.0,
                    )
                    sleep(0.1)
                    continue
                except Exception as e:
                    self.logger.error(f"could not initialize model {tb_str(e)}")
                    raise

        self.logger.debug("completed setup")

    def _add_path_topdown(self, top):
        if top in sys.path:
            return
        for root, _, __ in os.walk(top, topdown=True):
            sys.path.append(root)

    def _download_custom_files(self):
        # Check if there are any custom files and download them if so
        datasets = self.scenario.get("datasets", [])
        if len(datasets) > 0:
            self.nc.get_download_files(datasets)

    def instantiate_model(self):
        self.numerous_job.system = self

        for component in self.unrolled_components:
            for constant in component.constants.keys():
                val = component.constants[constant]
                # Handle references
                if isinstance(val, dict) and "ref" in val:
                    if val["ref"] in self.references:
                        ref_val = self.references[val["ref"]]
                    else:
                        ref_val = None
                        self.logger.warning(
                            f"{val['ref']} not found for {component.name} with {constant}. "
                            f"Available references are {self.references.keys()}"
                        )

                    component.constants[constant] = ref_val

            component.model = component._initialize(self, self.wrapper)

            # Add newly create component model to references to be passed in for subsequent ref values
            self.references[component.component_id] = component.model
            if component.container_id:
                self.references[component.container_id] = component.model

    def _dynamic_import_components(self):
        for component in self.unrolled_components:
            self._add_path_topdown(component._folder)
            component._dynamicimport()

    def update_inputs(self, data):
        if data is None:
            return
        component_data = copy.deepcopy(data)
        component_data.pop("_index")
        component_data.pop("_index_relative")
        self.input_data = component_data

        for input_tag, value in component_data.items():
            try:
                component_name = input_tag.split(".")[0]
                subcomponents = input_tag.split(".")[1:]
            except IndexError:
                logger.error(
                    f"input {input_tag} could not be split into component and subcomponent"
                )
                raise

            if component_name in self.components:
                self.components[component_name].inputs.update(
                    {".".join(subcomponents): value}
                )

    def _format_job_spec(self, scenario):
        simcomponents = scenario["simComponents"]
        component_list = []
        components = {}

        for simcomponent in simcomponents:
            if simcomponent.get("disabled", False) or simcomponent.get(
                "isSubcomponent"
            ):
                continue
            component = self._add_components_recursive(
                None, simcomponents, simcomponent
            )
            components.update({simcomponent["name"]: component})
            component_list.append(component)

        return component_list

    def _find_subcomponents(self, subcomps, simcomponents):
        subcomponents = []
        for subcomp in subcomps:
            for component in simcomponents:
                if component.get("uuid") == subcomp.get("uuid"):
                    subcomponents.append(component)
        return subcomponents

    def _add_components_recursive(
        self, parent: Optional[Component], simcomponents, simcomponent, level=0
    ):
        component = self._add_component(simcomponent, parent)

        subcomps = simcomponent.get("subcomponents")
        if subcomps:
            subcomponents_list = self._find_subcomponents(subcomps, simcomponents)
            for subcomponent in subcomponents_list:
                if subcomponent.get("disabled", False):
                    continue
                subcomponent = self._add_components_recursive(
                    component, simcomponents, subcomponent, level=level + 1
                )
                component._add_subcomponent(subcomponent)
        return component

    def _add_component(self, simcomponent, parent):
        handle = None
        has_model = False
        folder = "."
        path = "."

        # Get values for helper variables
        item_class = simcomponent["item_class"]
        items = []
        if item_class:
            items = item_class.split(".")
        component_name = simcomponent["name"]
        component_id = simcomponent["uuid"]
        container_id = (
            simcomponent["containerID"] if "containerID" in simcomponent else None
        )

        if item_class and len(items) > 1:
            has_model = True

            model_file_name = items[-2]
            handle = items[-1]

            if len(items) > 2:
                model_name = items[0]

                folder = f"{self._model_folder}/{model_name}"
                path = f"{self._model_folder}.{model_name}.{model_file_name}"
            else:
                folder = f"{self._model_folder}"
                path = f"{self._model_folder}.{model_file_name}"
        else:
            self.logger.warning(
                f"Skipping adding model for {component_name} as it does not obey requirements: "
                f"item_class is not None and must be formatted as x.y/x.y.z "
                f"(item_class: {item_class})"
            )

        component_constants = {}
        if simcomponent["parameters"]:
            for d in simcomponent["parameters"]:
                if d["type"] == "file":
                    component_constants[d["id"]] = d["value"]["name"]
                elif d["type"] == "reference":
                    component_constants[d["id"]] = {"ref": d["value"]}
                else:
                    component_constants[d["id"]] = d["value"]

        # Format parameters ('inputVariables' in frontend)
        component_parameters = (
            {
                d["id"]: f'{simcomponent["name"]}.{d["id"]}'
                for d in simcomponent["inputVariables"]
            }
            if simcomponent["inputVariables"]
            else {}
        )

        # Make a complete dict with formatted specifications

        if self._model_folder == ".":
            path = path.lstrip(".")

        component = Component(
            component_id=component_id,
            container_id=container_id,
            name=component_name,
            folder=folder,
            path=path,
            handle=handle,
            constants=component_constants,
            parameters=component_parameters,
            import_flag=has_model and self._model_folder is not None,
            item_class=item_class,
            parent=parent,
        )

        return component



import pkg_resources

from ..utils import bold, green


def command_version():
    version = pkg_resources.get_distribution("numerous.sdk").version
    print(green(f"{bold('numerous.sdk')} command line tool v{bold(version)}"))



import datetime
import logging
import os
import signal
from importlib.metadata import version
from threading import Event, Thread
from time import sleep
from typing import Optional

from numerous.cli.client import get_client
from numerous.client import NumerousClient
from numerous.client.numerous_client import ScenarioStatus

from .errors import tb_str
from .job import ExitCode, NumerousBaseJob

VERSION = version("numerous.sdk")
APPVERSION = os.getenv("APPVERSION", "?")

local = os.getenv("KUBERNETES_SERVICE_HOST", "") == ""


class NumerousApp:
    def __init__(
        self,
        numerous_job: NumerousBaseJob,
        appname="defaultnumerousApp",
        max_restarts=0,
        model_folder: Optional[str] = None,
        working_folder: Optional[str] = None,
        reset_job: bool = False,
        trace: bool = True,
    ):
        log_level = os.getenv("LOG_LEVEL", logging.root.level)

        self.appname = appname

        self.backup = Event()
        self.terminate = Event()
        self.scenario = None
        self.model_definition = None
        self.files = None
        self.job_spec = None
        self.solver_settings = None
        self.sim_job = None
        self.subscribe = False
        self.solver_settings = None
        self.logger = None
        self.allow_hibernation = False
        self.start_time = None
        self.end_time = None
        self.scenarioname = None
        self.max_restarts = max_restarts
        self.status = 0
        self.logger = logging.getLogger(self.appname)
        self.logger.setLevel(level=log_level)
        self.model_folder = model_folder
        self.numerous_job = numerous_job
        self.result = None
        self.working_folder = working_folder
        if local:
            self.client = get_client(trace=trace)
        else:
            self.client = NumerousClient(trace=trace)

        self.client.logger.setLevel(log_level)

        class __Output:
            def __init__(self, client: NumerousClient, buffer_size: int):
                self._output = client.new_writer(buffer_size=buffer_size)

            def write_row(self, t, row):
                self._output.write_row(row)

            def close(self):
                self._output.close()

        self.output = __Output(self.client, buffer_size=100)
        self.setup_scenario()

    def setup_scenario(self):
        self.client.set_scenario_progress(
            "initializing", ScenarioStatus.ENVIRONMENT_INITIALIZING, 0.0, force=True
        )

        self.scenario, self.model_definition, self.files = self.client.get_scenario()

        main_job = None
        main_job_id = None
        for job in self.scenario.get("jobs", {}).values():
            if job.get("isMain", True):
                main_job = job
                break
        org_job_id = self.client._job_id
        if main_job:
            main_job_id = main_job.get("id")
            self.client._job_id = main_job_id
        else:
            self.logger.warning("no main job found on scenario")

        self.client._run_settings = None
        # Prepare simulation setup

        self.start_time = self.client.run_settings.start.timestamp()
        self.end_time = self.client.run_settings.end.timestamp()
        subscribe = False
        if main_job_id:
            subscribe = (
                self.scenario["jobs"][main_job_id]
                .get("runSettings", {})
                .get("runMode", "duration")
                == "continuous"
            )
        self.subscribe = subscribe
        self.scenarioname = self.scenario["scenarioName"]
        self.allow_hibernation = self.client.params.get("allow_hibernation", False)
        self.logger.info(f"allow hibernation: {self.allow_hibernation}")
        self.logger.info(f"subscribe {self.subscribe}")
        self.client._job_id = org_job_id

    def loop(self):
        self.logger.info(
            f"starting numerous app for {self.appname} on scenario {self.scenarioname}"
        )
        thread = Thread(target=self.numerous_job._start, args=(self,), daemon=False)
        thread.start()
        starttime = datetime.datetime.now()
        last_checkpoint = starttime
        restarts = 0
        warned = False

        while not self.terminate.is_set():
            checkpoint_time = datetime.datetime.now()
            if (
                not thread.is_alive()
                and self.numerous_job._exit_code.value <= ExitCode.ERROR.value
            ):
                if not warned:
                    self.client.set_scenario_progress(
                        message=f"{self.numerous_job._exit_code.description}. Retrying after 5 mins "
                        f"({restarts + 1}/{self.max_restarts})",
                        status=ScenarioStatus.RUNNING,
                        force=True,
                    )
                    warned = True
                if (
                    datetime.datetime.now() - starttime
                    > datetime.timedelta(seconds=300)
                ) and restarts < self.max_restarts:
                    self.logger.error(f"{self.appname} is dead.. restarting")
                    self.status = 0
                    thread = Thread(
                        target=self.numerous_job._start, args=(self,), daemon=False
                    )
                    thread.start()
                    sleep(5)
                    restarts += 1
                    starttime = datetime.datetime.now()
                    warned = False

                elif restarts >= self.max_restarts:
                    self.logger.error(
                        f"{self.appname} is dead and reached max restarts"
                    )
                    break

            elif not thread.is_alive():
                self.logger.debug(f"thread has ended with code {self.status}")
                break

            if checkpoint(checkpoint_time, last_checkpoint):
                self.backup.set()
                last_checkpoint = checkpoint_time

            sleep(0.5)
        self.logger.debug(f"waiting for {self.appname} to shut down...")
        thread.join()
        self.logger.info(f"{self.appname} stopped")
        self.logger.info("Main thread stopped gracefully")

        return self.result

    def _run(self):
        """
        Runs the simulation job with the model located in model_folder_name. Handles all events.
        """
        log_level = os.getenv("LOG_LEVEL", logging.root.level)
        logger = logging.getLogger("numerous-image-tools")
        logger.setLevel(level=log_level)
        terminate = Event()

        def handler(sig, frame):
            logger.debug(f"main thread received {sig} {frame}")
            terminate.set()

        signal.signal(signal.SIGINT, handler)
        signal.signal(signal.SIGTERM, handler)

        try:
            if self.numerous_job is None:
                raise KeyError("no simulation job specified")
            if self.appname is None:
                logger.warning("no appname specified")

            logger.info(
                f"Welcome to numerous image tools. Base version {VERSION} - app {self.appname}"
            )

            #  Create a temporary working folder, that can be used to place generated and downloaded files.
            os.makedirs(self.working_folder, exist_ok=True)

            self.terminate = terminate
            self.client.terminate_event = terminate
            self.loop()

            self.client.set_scenario_progress(
                self.numerous_job._exit_code.description,
                self.numerous_job._exit_code.scenariostatus,
                force=True,
            )

            logger.debug(f"setting state {self.numerous_job._exit_code}")
            self.client.state.set("_status", self.numerous_job._exit_code.name)

        except Exception as e:
            logger.error(f"unhandled exception: {tb_str(e)}")
            self.client.set_scenario_progress(
                "unhandled error", ScenarioStatus.FAILED, 0.0, force=True
            )
            logger.debug("setting state -2")
            self.client.state.set("_status", -2)
        finally:
            self.output.close()
            self.client.close()
            logger.info("application ended")
            return self.result


def checkpoint(tnow, tlast, interval=60):
    dt = datetime.timedelta(minutes=interval)
    if tnow - tlast > dt:
        return True
    else:
        return False


def run_job(
    numerous_job=None,
    appname=None,
    max_restarts=0,
    model_folder="models",
    working_folder="tmp",
    reset_job=None,
    trace: bool = True,
):
    """
    Instantiates the application and runs it inside the try/except
    """

    app = NumerousApp(
        appname=appname,
        max_restarts=max_restarts,
        numerous_job=numerous_job,
        model_folder=model_folder,
        working_folder=working_folder,
        reset_job=reset_job,
        trace=trace,
    )

    app._run()



import json
import logging
from typing import Any, Iterable, Optional, TypedDict

from numerous.grpc import spm_pb2, spm_pb2_grpc

log = logging.getLogger(__name__)


class ScenarioSetting(TypedDict):
    path: list[str]
    type: str
    value: float
    id: str
    display_name: str


class OptimizationSetting:
    def __init__(self, path: list[str], setting_type: str, value: Any, raw: Any):
        self.path = path
        self.type = setting_type
        self.value = value
        self.raw = raw

    def scenario_setting(self, value: float) -> ScenarioSetting:
        return {
            "id": self.raw.get("id"),
            "display_name": self.raw.get("displayName"),
            "type": self.type,
            "value": value,
            "path": self.path,
        }


def extract_component(components: dict[str, Any], component: dict[str, Any]):
    value: dict[str, dict[str, Any]] = {"params": {}, "components": {}}
    for param in component.get("parameters", []):
        value["params"][param["id"]] = param.get("value")
    for subcomponent_ref in component.get("subcomponents", []):
        subcomponent = component.get(subcomponent_ref["uuid"], {})
        value["components"][subcomponent["name"]] = extract_component(
            components, subcomponent
        )
    return value


class OptimizationConfiguration:
    def __init__(
        self, spm_stub: spm_pb2_grpc.SPMStub, project_id: str, scenario_id: str
    ):
        scenario_response = spm_stub.GetScenario(
            spm_pb2.Scenario(project=project_id, scenario=scenario_id)
        )
        scenario = json.loads(scenario_response.scenario_document)

        if scenario.get("optimize", False):
            raise RuntimeError(
                f"Optimization was not enabled on scenario {scenario_id} in project {project_id}."
            )

        target_scenario = self._load_target_scenario_data_and_set_job_id(
            spm_stub, project_id, scenario
        )
        self.settings = list(self._parse_optimization_settings(target_scenario))
        components = self._components_by_uuid(scenario)
        self._parse_aggregator_configuration(components)
        self._parse_goal_function_configuration(components)

    def _load_target_scenario_data_and_set_job_id(
        self, spm_stub: spm_pb2_grpc.SPMStub, project_id: str, scenario: dict[str, Any]
    ) -> dict:
        self.target_scenario_id = scenario["optimizationTargetScenarioID"]
        target_scenario_request = spm_pb2.Scenario(
            project=project_id, scenario=self.target_scenario_id
        )
        target_scenario_response = spm_stub.GetScenario(target_scenario_request)
        target_scenario = json.loads(target_scenario_response.scenario_document)

        if not isinstance(target_scenario, dict):
            raise TypeError(f"Received invalid target scenario data: {target_scenario}")

        self.target_job_id = next(
            job_id
            for job_id, job in target_scenario["jobs"].items()
            if job.get("isMain", False)
        )
        return target_scenario

    def _parse_optimization_settings(
        self, target_scenario: dict[str, Any]
    ) -> Iterable[OptimizationSetting]:
        setting_paths = target_scenario.get("optimizationSettingsPaths", {})
        setting_paths.pop("__CONVERTED_FROM_MAP", None)  # Remove helper attribute
        settings_components = target_scenario.get(
            "optimizationSettingsSimComponents", {}
        )
        settings_components.pop("__CONVERTED_FROM_MAP", None)  # Remove helper attribute
        for settings_component_uuid, settings_components in settings_components.items():
            [
                settings_component
            ] = settings_components  # Assumption that exactly one component exists
            setting_path = setting_paths.get(settings_component_uuid)
            setting_type = settings_component_uuid.split("_")[
                -1
            ]  # param, value, offset, scale
            setting_value = extract_component(settings_components, settings_component)
            component: Optional[dict[str, Any]] = next(
                (
                    c
                    for c in target_scenario.get("simComponents", [])
                    if c["uuid"] == setting_path[-2]
                ),
                None,
            )
            if component is None:
                log.warning("Invalid component for optimization: %s", setting_path[-2])
                continue
            component_key = (
                "parameters" if setting_type == "param" else "inputVariables"
            )
            setting_raw = next(
                (
                    attr
                    for attr in component[component_key]
                    if attr["uuid"] == setting_path[-1]
                ),
                None,
            )
            setting_raw = self._add_setting_attrs_from_scenario(
                target_scenario,
                component_key,
                setting_path,
                setting_raw,
            )
            if None in (setting_path, setting_type, setting_value):
                log.warning(
                    "Invalid optimization setting component: %s", settings_component
                )
                continue
            yield OptimizationSetting(
                setting_path, setting_type, setting_value, setting_raw
            )

    def _parse_aggregator_configuration(self, components: dict[str, dict[str, Any]]):
        self.aggregator_config = None
        aggregator_component_uuid = next(
            (
                comp["uuid"]
                for comp in components.values()
                if comp["name"] == "aggregator"
            ),
            None,
        )
        if aggregator_component_uuid not in components:
            raise RuntimeError("Missing required aggregator component")

        aggregator_component = components.get(aggregator_component_uuid, {})
        self.aggregator_config = extract_component(components, aggregator_component)

    def _parse_goal_function_configuration(self, components: dict[str, dict[str, Any]]):
        self.goal_function_inputs = []
        self.goal_function_config = None
        goal_function_component_uuid = next(
            (
                comp["uuid"]
                for comp in components.values()
                if comp["name"] == "goal_function"
            ),
            None,
        )
        if goal_function_component_uuid not in components:
            raise RuntimeError("Missing required goal function component")

        goal_function_component = components.get(goal_function_component_uuid, {})
        self.goal_function_config = extract_component(
            components, goal_function_component
        )
        for input_variable in goal_function_component.get("inputVariables", []):
            if input_variable["dataSourceType"] != "scenario":
                log.warning(
                    "Invalid input variable for optimization: %s", input_variable
                )
                continue
            self.goal_function_inputs.append(
                {
                    "tag": input_variable["tagSource"]["tag"],
                    "scale": input_variable["scaling"],
                    "offset": input_variable["offset"],
                }
            )

    @staticmethod
    def _components_by_uuid(scenario: dict[str, Any]) -> dict[str, dict[str, Any]]:
        return {comp["uuid"]: comp for comp in scenario.get("simComponents", [])}

    @staticmethod
    def _add_setting_attrs_from_scenario(
        scenario: dict, component_key: str, path: list[str], setting_raw: Optional[dict]
    ) -> Optional[dict[str, Any]]:
        attrs = ("id", "displayName")
        component: dict = next(
            (c for c in scenario.get("simComponents", []) if c["uuid"] == path[-2]), {}
        )
        setting: Optional[dict] = next(
            (s for s in component.get(component_key, []) if s["uuid"] == path[-1]), None
        )
        if setting is None:
            log.debug("Setting %s with key %s is not found", path, component_key)
        else:
            old_attrs = set() if setting_raw is None else setting_raw.keys()
            attrs_to_update = {
                attr: setting[attr]
                for attr in attrs
                if attr in setting and attr not in old_attrs
            }
            log.debug("Adding attrs %s of %s", attrs_to_update, setting_raw)
            if setting_raw is None:
                setting_raw = attrs_to_update if attrs_to_update else None
            else:
                setting_raw.update(**attrs_to_update)
        return setting_raw



import signal
from typing import Any, Callable, Optional

from numerous.sdk.connect.job_state import JobState


class HibernationHandler:
    def __init__(
        self, state_getter: Callable[[], JobState], last_read_key_setter: Callable
    ) -> None:
        self.hibernating = False
        self._state_getter = state_getter
        self._last_read_key_setter = last_read_key_setter
        self.callback: Optional[Callable[[], Any]] = None

    def hibernate(self) -> None:
        self.hibernating = True
        if self.callback is not None:
            self.callback()
        self._state_getter().commit()
        self._last_read_key_setter()
        signal.raise_signal(signal.SIGTERM)



from logging import Logger
from pathlib import Path
from typing import Any

from ..repository import NumerousRepository, RepositoryRemote
from ..utils import bold, cyan, print_load_warnings, yellow


def command_config(log: Logger, path: Path, ns: Any):
    path = Path.cwd() if path is None else path
    repo, load_warnings = NumerousRepository(path).load()
    print_load_warnings(load_warnings)

    name = path.name if repo.remote is None else repo.remote.name

    if all(var is None for var in [ns.remote, ns.organization, ns.user, ns.snapshot]):
        print(cyan(f"Current configuration for {bold(name)}"))
        print(f"remote:       {bold(repo.remote or '')}")
        print(f"snapshot:     {bold(repo.snapshot or '')}")
        print(f"scenario:     {bold(repo.scenario or '')}")
    else:
        print(cyan(f"Set configuration for {bold(name)}"))
        if ns.remote:
            print(f"remote       = {bold(ns.remote)}")
            repo.remote = RepositoryRemote(ns.remote)
        if ns.snapshot:
            print(f"snapshot     = {bold(ns.snapshot)}")
            repo.snapshot = ns.snapshot
        if ns.scenario:
            print(f"scenario     = {bold(ns.scenario)}")
            repo.scenario = ns.scenario
        if input(yellow("Save changes? (y/n) ")).strip().lower() == "y":  # nosec
            repo.save()



from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Generator, Optional, Union


class RunMode(Enum):
    CONTINUOUS = "continuous"
    DURATION = "duration"


@dataclass
class JobTime:
    """Contains information about the job runtime."""

    run_mode: RunMode
    start: datetime
    end: Optional[datetime]
    duration: Optional[timedelta]

    @staticmethod
    def from_document(job_data: dict[str, Any]) -> "JobTime":
        run_settings: dict[str, Any] = job_data["runSettings"]
        run_mode = RunMode(run_settings["runMode"])
        start = datetime.fromisoformat(run_settings["startDate"])
        end = (
            datetime.fromisoformat(run_settings["endDate"])
            if run_settings.get("endDate") is not None
            else None
        )
        return JobTime(
            run_mode=run_mode,
            start=start,
            end=end,
            duration=None if end is None else end - start,
        )

    def range(
        self, step: Union[timedelta, float] = 1.0, start: Optional[datetime] = None
    ) -> Generator[datetime, None, None]:
        """Range over the runtime of the job, from :attr:`start` to :attr:`end`,
        inclusive. If :attr:`end` is not defined, range forever.

        :param step: The step in time between yielded time value. If it is a `float` it
            is interpreted as seconds.
        :param start: If specified, it is used instead of the job start time.
        """
        step_ = step if isinstance(step, timedelta) else timedelta(seconds=step)
        t = start if start is not None else self.start
        while self.end is None or t < self.end:
            yield t
            t += step_



import sys
from logging import Logger
from pathlib import Path

from ..auth import AuthenticationError, login
from ..repository import NumerousRepository
from ..utils import bold, green, print_load_warnings, red


def command_login(log: Logger, path: Path):
    path = path or Path.cwd()

    try:
        repo, load_warnings = NumerousRepository(path).load()
        print_load_warnings(load_warnings)
    except Exception as excp:
        log.debug(f"Exception {type(excp)} {excp}")
        print(red(f"Cannot login: {bold(str(path))} is not a repository."))
        sys.exit(1)

    if repo.remote is None:
        print(red("Cannot login: No remote is configured for the repository."))
        print(
            f"Use the command {bold('numerous config --remote <REMOTE_URL>')} to configure a remote for the "
            "repository."
        )
        sys.exit(1)

    if repo.scenario is None:
        print(red("Cannot re-login: A scenario must be checked out."))
        print(f"Check out a scenario with {bold('numerous checkout <SCENARIO ID>')}")
        sys.exit(1)

    try:
        repo.refresh_token = None
        login(repo)
        print(green("Login succeeded."))
    except AuthenticationError:
        print(red("Login failed."))
        sys.exit(1)



import logging
import random
import string
import threading
from collections import deque
from enum import Enum
from functools import wraps
from time import time
from typing import Any, Callable, Iterator, Optional, Sequence, Union
from urllib.parse import urlparse

import grpc

from numerous.client_common.validation_interceptor import ValidationInterceptor

from . import config
from .config import GRPC_MAX_MESSAGE_SIZE

log = logging.getLogger("numerous.client")


class Interp(Enum):
    zero = 0
    linear = 1
    quadratic = 2


class JobStatus(Enum):
    ready = 0
    running = 1
    finished = 2
    request_termination = 3
    terminated = 4
    failed = 5
    requested = 6
    initializing = 7


class RepeatedFunction:
    def __init__(
        self, interval: float, function: Callable, run_initially=False, *args, **kwargs
    ) -> None:
        self._timer: Optional[threading.Timer] = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.next_call = time()

        if run_initially:
            self.function(*self.args, **self.kwargs)

    def _run(self) -> None:
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self) -> None:
        if not self.is_running:
            self.next_call += self.interval
            self._timer = threading.Timer(self.next_call - time(), self._run)
            self._timer.start()
            self.is_running = True

    def stop(self) -> None:
        if self._timer is not None:
            self._timer.cancel()
        self.is_running = False


class RestorableIterator(Iterator):
    """
    Returns items not handled items again after restore.
    >>> restorable_iterator = RestorableIterator(iter(range(4)))
    >>> assert list(restorable_iterator) == [0, 1, 2, 3]
    >>> restorable_iterator.restore()
    >>> assert list(restorable_iterator) == [0, 1, 2, 3]
    >>> restorable_iterator.restore()
    >>> assert list(next(restorable_iterator) for i in range(3)) == [0, 1, 2]
    >>> restorable_iterator.item_handled()
    >>> restorable_iterator.item_handled()
    >>> restorable_iterator.restore()
    >>> assert list(restorable_iterator) == [2, 3]
    >>> restorable_iterator.item_handled()
    >>> restorable_iterator.restore()
    >>> assert list(restorable_iterator) == [3]
    >>> assert list(restorable_iterator) == []
    """

    def __init__(self, data: Iterator):
        self._not_handled: deque = deque()
        self._restored_data: deque = deque()
        self._data = data
        self._lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self._lock:
            next_item = (
                self._restored_data.popleft()
                if self._restored_data
                else next(self._data)
            )
            self._not_handled.append(next_item)
            return next_item

    def item_handled(self):
        with self._lock:
            self._not_handled.popleft()

    def restore(self):
        with self._lock:
            self._restored_data = self._not_handled + self._restored_data
            self._not_handled = deque()


def initialize_grpc_channel(
    refresh_token: str,
    token_callback: Callable[[], str],
    server: Optional[str] = None,
    port: Optional[int] = None,
    secure: Optional[bool] = None,
    instance_id: Optional[str] = None,
):
    secure_channel = secure if secure is not None else not config.FORCE_INSECURE
    if secure_channel is None:
        raise RuntimeError("Secure channel must be specified")
    server = server or config.NUMEROUS_API_SERVER
    if server is None:
        raise RuntimeError("Server must be specified")
    port = port or config.NUMEROUS_API_PORT
    if port is None:
        raise RuntimeError("Port must be specified")

    log.debug("Initializing gRPC channel %s:%s", server, port)

    options = [
        ("grpc.max_message_length", GRPC_MAX_MESSAGE_SIZE),
        ("grpc.max_send_message_length", GRPC_MAX_MESSAGE_SIZE),
        ("grpc.max_receive_message_length", GRPC_MAX_MESSAGE_SIZE),
    ]

    if secure_channel:
        creds = grpc.ssl_channel_credentials()
        channel = grpc.secure_channel(f"{server}:{port}", creds, options)
    else:
        channel = grpc.insecure_channel(f"{server}:{port}", options)

    vi = ValidationInterceptor(
        token=refresh_token, token_callback=token_callback, instance=instance_id
    )
    channel = grpc.intercept_channel(channel, vi)
    return channel, vi.instance


def parse_api_url(url: str):
    parsed_url = urlparse(url)
    server = parsed_url.hostname
    port = parsed_url.port or (443 if parsed_url.scheme == "https" else 80)
    secure = parsed_url.scheme == "https"
    return server, port, secure


class InterpolationOutOfRangeError(Exception):
    def __init__(self, min_: float, max_: float, values: Sequence[float]):
        oor_values = [value for value in values if value < min_ or value > max_]
        message = f"{oor_values} out of range [{min_}-{max_}]"
        super().__init__(message)


DataValue = Union[int, float]
DataList = Sequence[DataValue]


def inrange(x: DataList):
    def outer(fun: Callable[[DataList], DataList]):
        def inner(x_data: Union[DataList, DataValue]) -> Sequence[float]:
            if isinstance(x_data, (int, float)):
                x_data = [float(x_data)]
            if any(min(x) > x_elem or max(x) < x_elem for x_elem in x_data):
                raise InterpolationOutOfRangeError(min(x), max(x), x_data)
            return fun(x_data)

        return inner

    return outer


def interp_zero(
    x: DataList,
    y: DataList,
    x_data: DataList,
) -> DataList:
    return [y[_get_source_index(x, x_data_elem)] for x_data_elem in x_data]


def _get_source_index(
    x: DataList,
    x_data_elem: DataValue,
):
    return [i for i, x_source_elem in enumerate(x) if x_source_elem <= x_data_elem][-1]


def interp_linear_coefficients(
    x: DataList,
    y: DataList,
) -> tuple[float, float]:
    x_mean = sum(x) / len(x)
    y_mean = sum(y) / len(y)

    ss_xy = sum((x_elem - x_mean) * (y_elem - y_mean) for x_elem, y_elem in zip(x, y))
    ss_xx = sum((x_elem - x_mean) ** 2 for x_elem in x)

    b1 = ss_xy / ss_xx
    b0 = y_mean - b1 * x_mean

    return b0, b1


def interp_linear(
    b0: float,
    b1: float,
    x_data: DataList,
) -> list:
    return [b1 * x_elem + b0 for x_elem in x_data]


def interp1d(
    x: DataList,
    y: DataList,
    kind: str = "zero",
) -> Callable[[Union[DataList, DataValue]], DataList]:
    if len(x) != len(y):
        raise ValueError("lengths of vectors must match")

    if kind == "zero":

        @inrange(x)
        def wrapper_zero(xi: DataList):
            return interp_zero(x, y, xi)

        return wrapper_zero
    elif kind == "linear":
        b0, b1 = interp_linear_coefficients(x, y)

        @inrange(x)
        def wrapper_linear(xi: list[Union[int, float]]):
            return interp_linear(b0, b1, xi)

        return wrapper_linear

    else:
        raise NotImplementedError(f"method {kind} not implemented")


def _capped_repr(value: str, max_length: int = 100) -> str:
    value_repr = repr(value)
    capped_repr = value_repr[:max_length]
    if len(value_repr) > len(capped_repr):
        return f"{capped_repr}..."
    else:
        return capped_repr


def trace(func: Callable):
    funcname = func.__name__

    @wraps(func)
    def wrapper(self, *args: Any, **kwargs: Any):
        if (
            not getattr(self, "trace", None)
            or not self.trace
            or not getattr(self, "logger", None)
        ):
            return func(self, *args, **kwargs)

        all_args = [_capped_repr(arg) for arg in args] + [
            f"{key}={value!r}" for key, value in kwargs.items()
        ]
        call_id = "".join(random.sample(string.hexdigits, k=4))

        log.debug(
            "%s:called:%s.%s(%s)",
            call_id,
            type(self).__name__,
            funcname,
            ", ".join(all_args),
        )
        try:
            value = func(self, *args, **kwargs)
            log.debug(
                "%s:return:%s.%s -> %s",
                call_id,
                type(self).__name__,
                funcname,
                _capped_repr(value),
            )
        except Exception as e:
            log.debug(
                "%s:except:%s.%s -> %s(%s)",
                call_id,
                type(self).__name__,
                funcname,
                type(e).__name__,
                e,
            )
            raise

        return value

    return wrapper



from time import sleep
from typing import Any, Callable, Generator, Iterable


class RequestResponseStream:
    # TODO: Implement async methods
    def __init__(
        self, stream_endpoint: Callable[[Iterable], Generator], check_delay: float = 0.1
    ):
        self._check_delay = check_delay

        self._request: Any = None
        self._response_iterator = stream_endpoint(self)

    def send(self, request):
        while self._request is not None:
            sleep(self._check_delay)
        self._request = request
        return next(self._response_iterator)

    def close(self):
        while self._request is not None:
            sleep(self._check_delay)
        self._request = StopIteration()

    def __iter__(self):
        return self

    def __next__(self):
        while self._request is None:
            sleep(self._check_delay)

        if type(self._request) is StopIteration:
            self._request = None
            raise StopIteration()

        request = self._request
        self._request = None
        return request

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # noqa: F841
        self.close()



from time import sleep
from typing import Callable, Generator, Generic, Iterable, Optional, TypeVar

RequestType = TypeVar("RequestType")
ResponseType = TypeVar("ResponseType")


class RequestResponseStream(Generic[RequestType, ResponseType]):
    """A helper class for wrapping Stream-Stream gRPCs."""

    def __init__(
        self,
        stream_endpoint: Callable[
            [Iterable[RequestType]], Generator[ResponseType, None, None]
        ],
        check_interval_secs: float = 0.1,
    ):
        """Initialize the :class:`RequestResponseStream`.

        :param stream_endpoint: The endpoint the :class:`RequestResponseStream` should wrap.
        :param check_interval_secs: The time interval in seconds, for checking whether
        the current request is sent, or the :class:`RequestResponseStream` is closed.
        """
        self._check_interval = check_interval_secs
        self._request: Optional[RequestType] = None
        self._closed: bool = False
        self._response_iterator = stream_endpoint(self)

    def _await_request_sent(self):
        while self._request is not None:
            sleep(self._check_interval)

    def send(self, request: RequestType) -> ResponseType:
        """Sends a request after waiting for the previous request to have been sent, if
        there is one.

        :param request: The request to send.
        :return: The response.
        """
        self._await_request_sent()
        self._request = request
        return next(self._response_iterator)

    def close(self) -> None:
        """Close the :class:`RequestResponseStream` blocking until any outstanding
        request has been delivered.
        """
        self._await_request_sent()
        self._request = None
        self._closed = True

    def __iter__(self) -> "RequestResponseStream[RequestType, ResponseType]":
        return self

    def __next__(self) -> RequestType:
        while not self._closed and self._request is None:
            sleep(self._check_interval)

        if self._closed:
            raise StopIteration()
        else:
            request = self._request
            self._request = None
            return request  # type: ignore[return-value]



"""Contains the implementation of the job client."""

import json
import logging
import os
import signal
import sys
import threading
from types import FrameType, TracebackType
from typing import Any, Callable, Optional
from urllib.parse import urlparse

import grpc
from numerous.grpc import spm_pb2
from numerous.grpc.spm_pb2_grpc import SPMStub, TokenManagerStub

from numerous.cli.local import create_local_execution_from_repository
from numerous.sdk.connect.access_token import RefreshingAccessToken
from numerous.sdk.connect.auth import AccessTokenAuthMetadataPlugin
from numerous.sdk.connect.command_handler import CommandHandler
from numerous.sdk.connect.config import Config
from numerous.sdk.connect.file_manager import FileManager, FileUploadContext
from numerous.sdk.connect.hibernation import HibernationHandler
from numerous.sdk.connect.inputs import (
    HIBERNATE_SELF,
    HibernatingSourceAction,
    InputReader,
)
from numerous.sdk.connect.job_state import JobState
from numerous.sdk.connect.job_utils import JobIdentifier
from numerous.sdk.connect.reader import Reader
from numerous.sdk.connect.status_handler import JobStatus, StatusHandler
from numerous.sdk.connect.subscription import Subscription
from numerous.sdk.connect.writer import Writer
from numerous.sdk.models.job import Job
from numerous.sdk.models.job_time import JobTime
from numerous.sdk.models.optimization import OptimizationConfiguration
from numerous.sdk.models.parameter import Parameter
from numerous.sdk.models.scenario import Scenario


class JobLogHandler(logging.Handler):
    def __init__(
        self,
        spm: SPMStub,
        identity: JobIdentifier,
        execution_id: str,
        current_root_log_level: int = 0,
    ) -> None:
        super().__init__()
        self._spm_client = spm
        self._project_id = identity.project_id
        self._scenario_id = identity.scenario_id
        self._execution_id = execution_id
        self._previous_root_log_level = current_root_log_level

    def emit(self, record: logging.LogRecord) -> None:
        if logging.Logger.root.level > record.levelno:
            return

        self._spm_client.PushExecutionLogEntries(
            spm_pb2.LogEntries(
                execution_id=self._execution_id,
                log_entries=[self.format(record)],
                timestamps=[record.created],
                scenario=self._scenario_id,
                project_id=self._project_id,
            )
        )

    def get_previous_root_level(self) -> int:
        return self._previous_root_log_level


class JobClient:
    """The JobClient is the recommended way to connect to the numerous platform."""

    def __init__(
        self,
        channel: grpc.Channel,
        identity: JobIdentifier,
        execution_id: str,
        config: Optional[Config] = None,
    ) -> None:
        """Initialize the job client with a gRPC channel. The channel must be
        configured with credentials.

        :param channel: A gRPC channel configured with required authorization.
        :param identity: Contains identity information for the job object and related objects.
        """
        self._config = Config() if config is None else config
        self._channel = channel
        self._spm_client = SPMStub(self._channel)
        self._identity = identity
        self._execution_id = execution_id
        self._logger: logging.Logger = self._log_init()
        self._file_manager = FileManager(
            self._spm_client, self._identity.project_id, self._identity.scenario_id
        )
        self._job_state: Optional[JobState] = None
        self._job = None
        self._scenario_document: Optional[dict[str, Any]] = None
        self._scenario: Optional[Scenario] = None
        self._reader: Optional[Reader] = None
        self._input_reader: Optional[InputReader] = None
        self._writer: Optional[Writer] = None
        self._terminate_callback: Optional[Callable] = None
        self._status_handler = StatusHandler(
            self._spm_client, self._identity, self._config
        )
        self._setup_sigterm_handling()

        def set_last_read_key() -> None:
            if self._input_reader:
                self._input_reader.store_last_read_key()

        self._hibernation_handler = HibernationHandler(
            lambda: self.state, set_last_read_key
        )
        self._command_handler = self._create_command_handler(self._hibernation_handler)

    def _setup_sigterm_handling(self) -> None:
        def _sigterm_handler(signalnum: int, frame: Optional[FrameType]):  # noqa: F841
            if self._hibernation_handler.hibernating:
                sys.exit(0)

            if self._terminate_callback:
                thread = threading.Thread(target=self._terminate_callback, daemon=True)
                thread.start()
                thread.join(timeout=self._config.terminate_handler_timeout)
            sys.exit(1)

        signal.signal(signal.SIGTERM, _sigterm_handler)

    def _create_command_handler(
        self, hibernating_handler: HibernationHandler
    ) -> CommandHandler:
        subscription = Subscription(
            self._identity.project_id,
            self._identity.scenario_id,
            self._spm_client,
            [
                ".".join(
                    (
                        "COMMAND",
                        self._identity.project_id,
                        self._identity.scenario_id,
                        self._identity.job_id,
                    )
                )
            ],
        )
        return CommandHandler(subscription, hibernating_handler)

    def _log_init(self) -> logging.Logger:
        """Job Client Log Initialization"""
        job_client_logger = logging.getLogger("job_client")
        job_client_logger.setLevel(logging.INFO)
        self._add_job_log_handler_to_root()
        return job_client_logger

    def _add_job_log_handler_to_root(self):
        root_level = logging.root.level
        if not logging.root.hasHandlers():
            logging.root.setLevel(logging.INFO)
        job_log_handler = JobLogHandler(
            self._spm_client, self._identity, self._execution_id, root_level
        )
        job_log_handler.setFormatter(logging.Formatter(fmt=logging.BASIC_FORMAT))
        logging.root.addHandler(job_log_handler)

    def _remove_job_log_handler_from_root(self):
        job_log_handler = next(
            (item for item in logging.root.handlers if isinstance(item, JobLogHandler)),
            None,
        )
        if job_log_handler is not None:
            logging.root.handlers.remove(job_log_handler)
            logging.root.setLevel(job_log_handler.get_previous_root_level())

    def __enter__(self) -> "JobClient":
        """Return itself upon entering the context manager."""
        self.status = JobStatus.RUNNING
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],  # noqa: F841
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],  # noqa: F841
    ) -> Optional[bool]:
        """Closes the gRPC channel upon exiting the context manager."""
        self.close(exc_value)
        return None

    @staticmethod
    def connect(
        status: JobStatus = JobStatus.RUNNING, message: Optional[str] = None
    ) -> "JobClient":
        """Create a JobClient from environment variables.

        Uses the following environment variables:
         - `NUMEROUS_API_SERVER`
         - `NUMEROUS_API_PORT`
         - `NUMEROUS_API_REFRESH_TOKEN`
         - `NUMEROUS_PROJECT`
         - `NUMEROUS_SCENARIO`
         - `JOB_ID`
         - `NUMEROUS_EXECUTION_ID`
        """
        try:
            job_client = JobClient._from_environment()
        except KeyError:
            job_client = JobClient._from_numerous_repository()
        job_client.status = status
        job_client.message = message or "Running"
        return job_client

    @staticmethod
    def from_connection_params(
        hostname: str,
        port: str,
        refresh_token: str,
        identity: JobIdentifier,
        execution_id: str,
        config: Config,
    ) -> "JobClient":
        """Create a JobClient from connection parameters.

        :param hostname: Hostname of the numerous server
        :param port: gRPC port of the numerous server
        :param refresh_token: Refresh token for the execution.
        :param identity: Contains identity information for the job object and related objects.
        """
        token_manager = TokenManagerStub(
            grpc.secure_channel(
                f"{hostname}:{port}",
                grpc.ssl_channel_credentials(),
                JobClient.channel_options(config),
            )
        )

        authorized_channel = grpc.secure_channel(
            f"{hostname}:{port}",
            grpc.composite_channel_credentials(
                grpc.ssl_channel_credentials(),
                grpc.metadata_call_credentials(
                    AccessTokenAuthMetadataPlugin(
                        RefreshingAccessToken(token_manager, refresh_token)
                    )
                ),
            ),
            JobClient.channel_options(config),
        )

        return JobClient(authorized_channel, identity, execution_id, config)

    @staticmethod
    def channel_options(config: Config) -> list[tuple[str, Any]]:
        """Returns the default gRPC channel options."""
        return [
            ("grpc.max_message_length", config.grpc_max_message_size),
            ("grpc.max_send_message_length", config.grpc_max_message_size),
            ("grpc.max_receive_message_length", config.grpc_max_message_size),
        ]

    @staticmethod
    def _from_environment() -> "JobClient":
        """Create a JobClient from environment variables.

        Uses the following environment variables:
         - `NUMEROUS_API_SERVER`
         - `NUMEROUS_API_PORT`
         - `NUMEROUS_API_REFRESH_TOKEN`
         - `NUMEROUS_PROJECT`
         - `NUMEROUS_SCENARIO`
         - `JOB_ID`
         - `NUMEROUS_EXECUTION_ID`
        """
        return JobClient.from_connection_params(
            os.environ["NUMEROUS_API_SERVER"],
            os.environ["NUMEROUS_API_PORT"],
            os.environ["NUMEROUS_API_REFRESH_TOKEN"],
            JobIdentifier(
                os.environ["NUMEROUS_PROJECT"],
                os.environ["NUMEROUS_SCENARIO"],
                os.environ["JOB_ID"],
            ),
            os.environ["NUMEROUS_EXECUTION_ID"],
            config=Config.from_environment(),
        )

    @staticmethod
    def _from_legacy_client(client: Any) -> "JobClient":
        """Create a JobClient from a legacy :class:`~numerous.client.numerous_client.NumerousClient`.

        Connects to the same job as the legacy client is connected to.
        """
        from numerous.client import NumerousClient
        from numerous.client.common import initialize_grpc_channel

        if not isinstance(client, NumerousClient):
            raise TypeError(
                f"'client' argument must be of type NumerousClient, was {type(client).__name__}"
            )
        else:
            if client._refresh_token is None:
                raise RuntimeError("Legacy client does not have a refresh token")

            channel, _ = initialize_grpc_channel(
                refresh_token=client._refresh_token,
                token_callback=client._get_current_token,
                server=client._channel_server,
                port=client._channel_port,
                secure=client._channel_secure,
                instance_id=client._instance_id,
            )

            return JobClient(
                channel,
                JobIdentifier(client._project, client._scenario, client._job_id),
                client._execution_id,
                Config.from_environment(),
            )

    @staticmethod
    def _from_numerous_repository() -> "JobClient":
        repo, execution_id, refresh_token = create_local_execution_from_repository()
        if repo.remote is None:
            raise RuntimeError(
                f"Repository at {repo.path} does not have a remote configured."
            )
        if repo.scenario is None:
            raise RuntimeError(
                f"Repository at {repo.path} does not have a scenario checked out."
            )

        parsed_url = urlparse(repo.remote.api_url)
        port = parsed_url.port or (443 if parsed_url.scheme == "https" else 80)
        if parsed_url.hostname is None:
            raise RuntimeError(
                f"Repository at {repo.path} does not have a valid remote configured."
            )

        return JobClient.from_connection_params(
            parsed_url.hostname,
            str(port),
            refresh_token,
            JobIdentifier(
                repo.scenario.project_id,
                repo.scenario.id,
                repo.remote.job_id,
            ),
            execution_id,
            config=Config.from_environment(),
        )

    def open_output_file(self, file_name: str) -> FileUploadContext:
        """
        :param file_name: Output file name.
        :return: An object to be used under with statement.

        Example:
            >>> with JobClient.connect() as client:
            ...     with client.open_output_file("output_data") as output_file:
            ...         output_file.write("First file content")
            ...         output_file.write_from_file("first_file.csv")
            ...         output_file.write("Second file content")
            ...         output_file.write_from_file("second_file.csv")
        """
        return FileUploadContext(self._file_manager, file_name)

    def close(self, error: Optional[BaseException] = None) -> None:
        """Close the JobClient.

        Closes the JobClient's connection to the numerous platform, immediately
        terminating any active communication.

        :param error_occured: If `True`, sets the job status with `JobStatus.ERROR`.

        This method is idempotent.
        """
        if isinstance(error, SystemExit) and error.code == 0:
            error = None

        if error:
            self.log.error(
                "An error occured in the job. This may be due to an error in the job "
                "code. Please reach out to the developer with the error information "
                "below:",
                exc_info=error,
            )

        if self._writer is not None:
            self._writer.close(self._hibernation_handler.hibernating)

        self._set_completed_job_status(error is not None)
        self._remove_job_log_handler_from_root()
        self._complete_execution()
        self._status_handler.close()
        self._channel.close()
        self._command_handler.close()

    def _set_completed_job_status(self, error_occured: bool):
        if error_occured:
            self.status = JobStatus.FAILED
            self.message = "Error occured"
        elif self._hibernation_handler.hibernating:
            self.message = "Hibernating"
            self.status = JobStatus.HIBERNATING
        elif self.status not in (
            JobStatus.ERROR,
            JobStatus.FINISHED,
            JobStatus.STOPPED,
        ):
            self.message = "Finished"
            self.status = JobStatus.FINISHED

    def _complete_execution(self) -> None:
        """Signal to the back-end that the execution has ended."""
        self._spm_client.CompleteExecution(
            request=spm_pb2.CompleteExecutionMessage(
                execution=spm_pb2.ExecutionMessage(
                    project_id=self._identity.project_id,
                    scenario_id=self._identity.scenario_id,
                    job_id=self._identity.job_id,
                    execution_id=self._execution_id,
                ),
                hibernate=self._hibernation_handler.hibernating,
            )
        )

    @property
    def log(self) -> logging.Logger:
        """Return the Job Client logger for the job."""
        return self._logger

    @property
    def file_manager(self) -> FileManager:
        """Access the file manager of the job."""
        return self._file_manager

    @property
    def state(self) -> JobState:
        """The job state, which can be persisted across hibernations.

        It is a lazy property, that will load any remote state on access.
        """
        if self._job_state is None:
            self._job_state = JobState(
                self._spm_client, self._identity, self._execution_id
            )
        return self._job_state

    @property
    def parameters(self) -> dict[str, Parameter]:
        return self.job.parameters

    @property
    def scenario(self) -> Scenario:
        """The associated :class:`Scenario` configuration, with jobs, components,
        parameters and input variables.

        It is a lazy property, that will load the scenario data on access.
        """
        if self._scenario_document is None:
            self._scenario_document = self._get_scenario_document()

        if self._scenario is None:
            scenario_has_optimization_target = self._scenario_document.get(
                "optimizationTargetScenarioID", False
            )
            optimization = (
                OptimizationConfiguration(
                    self._spm_client,
                    self._identity.project_id,
                    self._identity.scenario_id,
                    self._scenario_document,
                )
                if scenario_has_optimization_target
                else None
            )

            self._scenario = Scenario.from_document(
                self._scenario_document, optimization
            )
        return self._scenario

    @property
    def job_time(self) -> JobTime:
        """The associated time setup of the job, which contains start and end times
        of the simulation, and the duration.
        """
        return self.job.time

    def _get_scenario_document(self) -> dict[str, Any]:
        if self._scenario_document is None:
            response = self._spm_client.GetScenario(
                spm_pb2.Scenario(
                    project=self._identity.project_id,
                    scenario=self._identity.scenario_id,
                )
            )
            self._scenario_document = json.loads(response.scenario_document)
        return self._scenario_document

    @property
    def status(self) -> JobStatus:
        """Status of the job, reported to the platform.

        Getting the status returns a locally cached value.
        """
        return self._status_handler.status

    @status.setter
    def status(self, value: JobStatus) -> None:
        self._status_handler.status = value

    @property
    def message(self) -> str:
        """Status message of the job, reported to the platform. Is truncated to at most 32 characters upon setting.

        Getting the status message returns a locally cached value.
        """
        return self._status_handler.message

    @message.setter
    def message(self, value: str):
        self._status_handler.message = value

    @property
    def progress(self) -> float:
        """Progress of the job, reported to the platform. Is clamped between 0.0 and 100.0 upon setting.

        Getting the progress returns a locally cached value.
        """
        return self._status_handler.progress

    @progress.setter
    def progress(self, value: float):
        self._status_handler.progress = value

    @property
    def reader(self) -> Reader:
        """Data reader for the job."""
        if self._reader is None:
            self._reader = Reader(
                self._spm_client,
                self._identity,
                self._execution_id,
            )
        return self._reader

    @property
    def writer(self) -> Writer:
        """Data writer for the job."""
        if self._writer is None:
            self._writer = Writer(
                self._spm_client,
                self._identity,
                self.job_time.start,
                self._execution_id,
                self._config.grpc_max_message_size,
                flush_margin_bytes=self._config.grpc_max_message_size // 8,
            )

        return self._writer

    def hibernate(self) -> None:
        """Hibernate the current job execution

        Terminates the Python process, committing the job state to the
        back-end. The job can be resumed again, and the state can be reloaded
        to continue.

        If the job is scheduled, it will then follow its schedule, and be
        resumed at the next scheduled resume.
        """
        self._hibernation_handler.hibernate()

    def set_terminate_callback(self, callback: Callable[[], Any]) -> None:
        """Set a custom function to be called if the job is terminated."""
        self._terminate_callback = callback

    def set_hibernate_callback(self, callback: Callable[[], Any]) -> None:
        """Set a custom function to be called if the job is hibernated."""
        self._hibernation_handler.callback = callback

    @property
    def job(self) -> Job:
        """The model of the current job."""
        return self.scenario.jobs[self._identity.job_id]

    def read_inputs(
        self,
        step: float,
        repeat_spacing: Optional[float] = None,
        hibernating_source_action: HibernatingSourceAction = HIBERNATE_SELF,
    ) -> InputReader:
        """Get an iterator over all input variables in the scenario.

        This method can only be called once in the job runtime.

        :param step: The distance (in time) between each read. Zero-order hold is used
            for values at steps that are not in the original data set.
        :param repeat_spacing: The distance (in time) between the last data point of the
            source data set, and the repeated first data point when repeating.

        Returns an :class:`~numerous.sdk.connect.inputs.InputReader`, that can be
        iterated over.

        Example:
            >>> result = 0
            >>> with JobClient.connect() as job_client:
            ...     for t, data in job_client.read_inputs(step=3600.0):
            ...          result += data["component_name.input_variable_id"]
        """

        if self._input_reader:
            raise RuntimeError("You can only read inputs once during a job runtime.")

        self._input_reader = InputReader(
            self._spm_client,
            hibernation_handler=self._hibernation_handler,
            identity=self._identity,
            execution_id=self._execution_id,
            job=self.job,
            input_sources=self.scenario.input_sources,
            input_variables=self.scenario.input_variables,
            step=step,
            repeat_spacing=repeat_spacing,
            hibernating_source_action=hibernating_source_action,
        )
        return self._input_reader



import logging
from pathlib import Path
from typing import Optional, Union

from numerous.cli.local import create_local_execution_from_repository
from numerous.client import NumerousClient


def get_client(
    path: Optional[Path] = None,
    clear_data: Optional[bool] = None,
    no_log: bool = False,
    log_level: Union[str, int] = logging.ERROR,
    trace: bool = False,
) -> NumerousClient:
    repo, execution_id, refresh_token = create_local_execution_from_repository(path)

    if repo.remote is None or repo.scenario is None:
        raise RuntimeError(
            "Repository not configured correctly, try checking out a scenario or re-initializing."
        )

    return NumerousClient(
        url=repo.remote.api_url,
        project=repo.scenario.project_id,
        scenario=repo.scenario.id,
        job_id=repo.remote.job_id,
        refresh_token=refresh_token,
        execution_id=execution_id,
        clear_data=clear_data,
        log_level=log_level,
        no_log=no_log,
        trace=trace,
    )



from dataclasses import dataclass
from typing import Any, Iterable, Optional

from numerous.sdk.models.input import (
    InputSource,
    InputVariable,
    get_input_sources_from_scenario_document,
)
from numerous.sdk.models.optimization import OptimizationConfiguration

from .component import Component, extract_components
from .job import Job
from .parameter import Parameter, get_parameters_dict


@dataclass
class Scenario:
    id: str
    name: str
    components: dict[str, Component]
    parameters: dict[str, Parameter]
    jobs: dict[str, Job]
    input_sources: list[InputSource]
    input_variables: list[InputVariable]
    optimization: Optional[OptimizationConfiguration] = None

    @staticmethod
    def from_document(
        data: dict[str, Any], optimization: Optional[OptimizationConfiguration] = None
    ) -> "Scenario":
        input_sources_mapping = get_input_sources_from_scenario_document(data)
        components = extract_components(data, input_sources_mapping)

        return Scenario(
            id=data["id"],
            name=data.get("scenarioName", ""),
            components=components,
            jobs={
                job_id: Job.from_document(job_id, job_data)
                for job_id, job_data in data.get("jobs", {}).items()
            },
            input_sources=list(input_sources_mapping.values()),
            input_variables=_collect_input_variables(components.values()),
            optimization=optimization,
            parameters=get_parameters_dict(data.get("parameters", [])),
        )


def _collect_input_variables(components: Iterable[Component]) -> list[InputVariable]:
    input_variables: list[InputVariable] = []
    for component in list(components):
        input_variables += list(component.input_variables.values())
        input_variables += _collect_input_variables(
            [
                subcomponent
                for subcomponent_category in component.components.values()
                for subcomponent in subcomponent_category
            ]
        )
    return input_variables
